#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import atexit
import queue
import signal
import threading
import traceback
import weakref

from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import context as n_ctx
from neutron_lib.db import api as db_api
from oslo_log import log as logging

from neutron.api.rpc.callbacks import events as rpc_events
from neutron.api.rpc.handlers import resources_rpc
from neutron.objects import address_group
from neutron.objects import network
from neutron.objects import ports
from neutron.objects import securitygroup
from neutron.objects import subnet

LOG = logging.getLogger(__name__)


def _setup_change_handlers_cleanup():
    atexit.register(_ObjectChangeHandler.clean_up)
    signal.signal(signal.SIGINT, _ObjectChangeHandler.clean_up)
    signal.signal(signal.SIGTERM, _ObjectChangeHandler.clean_up)


class _ObjectChangeHandler(object):
    MAX_IDLE_FOR = 1
    _TO_CLEAN = weakref.WeakSet()

    def __init__(self, resource, object_class, resource_push_api):
        self._resource = resource
        self._obj_class = object_class
        self._resource_push_api = resource_push_api
        self._resources_to_push = queue.Queue()
        self._semantic_warned = False
        for event in (events.AFTER_CREATE, events.AFTER_UPDATE,
                      events.AFTER_DELETE):
            registry.subscribe(self.handle_event, resource, event)

        self._stop = threading.Event()
        self._worker = threading.Thread(
            target=self.dispatch_events,
            name='ObjectChangeHandler[%s]' % self._resource,
            daemon=True)
        self._worker.start()
        self._TO_CLEAN.add(self)

    def stop(self):
        self._stop.set()

    def wait(self):
        """Waits for all outstanding events to be dispatched."""
        self._resources_to_push.join()

    def _is_session_semantic_violated(self, context, resource, event):
        """Return True and print an ugly error on transaction violation.

        This code is to print ugly errors when AFTER_CREATE/UPDATE
        event transaction semantics are violated by other parts of
        the code.
        """
        if not db_api.is_session_active(context.session):
            return False
        if not self._semantic_warned:
            stack = traceback.extract_stack()
            stack = "".join(traceback.format_list(stack))
            LOG.warning("This handler is supposed to handle AFTER "
                        "events, as in 'AFTER it's committed', "
                        "not BEFORE. Offending resource event: "
                        "%(r)s, %(e)s. Location:\n%(l)s",
                        {'r': resource, 'e': event, 'l': stack})
            self._semantic_warned = True
        return True

    def handle_event(self, resource, event, trigger, payload):
        """Callback handler for resource change that pushes change to RPC.

        We always retrieve the latest state and ignore what was in the
        payload to ensure that we don't get any stale data.
        """
        if self._is_session_semantic_violated(
                payload.context, resource, event):
            return
        resource_id = payload.resource_id
        # we preserve the context so we can trace a receive on the agent back
        # to the server-side event that triggered it
        self._resources_to_push.put((resource_id, payload.context.to_dict()))

    def dispatch_events(self):
        # TODO(kevinbenton): now that we are batching these, convert to a
        # single get_objects call for all of them
        LOG.debug('Thread %(name)s started', {'name': self._worker.name})
        while not self._stop.is_set():
            try:
                resource_id, context_dict = self._resources_to_push.get(
                    timeout=self.MAX_IDLE_FOR)
                context = n_ctx.Context.from_dict(context_dict)
                # attempt to get regardless of event type so concurrent delete
                # after create/update is the same code-path as a delete event
                with db_api.get_context_manager().independent.reader.using(
                        context):
                    obj = self._obj_class.get_object(context, id=resource_id)
                # CREATE events are always treated as UPDATE events to ensure
                # listeners are written to handle out-of-order messages
                if obj is None:
                    rpc_event = rpc_events.DELETED
                    # construct a fake object with the right ID so we can
                    # have a payload for the delete message.
                    obj = self._obj_class(id=resource_id)
                else:
                    rpc_event = rpc_events.UPDATED
                self._resource_push_api.push(context, [obj], rpc_event)
                self._resources_to_push.task_done()
            except queue.Empty:
                pass
            except Exception as e:
                LOG.exception(
                    "Exception while dispatching %(res)s events: %(e)s",
                    {'res': self._resource, 'e': e})
        LOG.debug('Thread %(name)s finished with %(msgs)s unsent messages',
                  {'name': self._worker.name,
                   'msgs': self._resources_to_push.unfinished_tasks})

    @classmethod
    def clean_up(cls, *args, **kwargs):
        """Ensure all threads that were created were destroyed cleanly."""
        while cls._TO_CLEAN:
            worker = cls._TO_CLEAN.pop()
            worker.stop()


class OVOServerRpcInterface(object):
    """ML2 server-side RPC interface.

    Generates RPC callback notifications on ML2 object changes.
    """

    def __init__(self, enable_signals=True):
        self._rpc_pusher = resources_rpc.ResourcesPushRpcApi()
        self._setup_change_handlers()
        # When running behind wsgi server (like apache2/mod_wsgi)
        # we should not register signals
        if enable_signals:
            _setup_change_handlers_cleanup()
        LOG.debug("ML2 OVO RPC backend initialized.")

    def _setup_change_handlers(self):
        """Setup all of the local callback listeners for resource changes."""
        resource_objclass_map = {
            resources.PORT: ports.Port,
            resources.SUBNET: subnet.Subnet,
            resources.NETWORK: network.Network,
            resources.SECURITY_GROUP: securitygroup.SecurityGroup,
            resources.SECURITY_GROUP_RULE: securitygroup.SecurityGroupRule,
            resources.ADDRESS_GROUP: address_group.AddressGroup,
        }
        self._resource_handlers = {
            res: _ObjectChangeHandler(res, obj_class, self._rpc_pusher)
            for res, obj_class in resource_objclass_map.items()
        }

    def wait(self):
        """Wait for all handlers to finish processing async events."""
        for handler in self._resource_handlers.values():
            handler.wait()



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib.db import api as db_api
from neutron_lib import exceptions as nlib_exc
from neutron_lib.plugins import directory
from oslo_db import exception as db_exc
from oslo_log import log
from oslo_utils import uuidutils
from sqlalchemy import or_
from sqlalchemy.orm import exc

from neutron._i18n import _
from neutron.db.models import securitygroup as sg_models
from neutron.db import models_v2
from neutron.objects import base as objects_base
from neutron.objects import ports as port_obj
from neutron.plugins.ml2 import models
from neutron.services.segments import db as seg_db
from neutron.services.segments import exceptions as seg_exc

LOG = log.getLogger(__name__)

# limit the number of port OR LIKE statements in one query
MAX_PORTS_PER_QUERY = 500


@db_api.CONTEXT_WRITER
def add_port_binding(context, port_id):
    record = models.PortBinding(
        port_id=port_id,
        vif_type=portbindings.VIF_TYPE_UNBOUND)
    context.session.add(record)
    return record


@db_api.CONTEXT_WRITER
def set_binding_levels(context, levels):
    if levels:
        for level in levels:
            level.create()
        LOG.debug("For port %(port_id)s, host %(host)s, "
                  "set binding levels %(levels)s",
                  {'port_id': levels[0].port_id,
                   'host': levels[0].host,
                   'levels': levels})
    else:
        LOG.debug("Attempted to set empty binding levels")


@db_api.CONTEXT_READER
def get_binding_level_objs(context, port_id, host):
    if host:
        pager = objects_base.Pager(sorts=[('level', True)])
        port_bl_objs = port_obj.PortBindingLevel.get_objects(
            context, _pager=pager, port_id=port_id, host=host)
        LOG.debug("For port %(port_id)s, host %(host)s, "
                  "got binding levels %(levels)s",
                  {'port_id': port_id,
                   'host': host,
                   'levels': port_bl_objs})
        return port_bl_objs


@db_api.CONTEXT_WRITER
def clear_binding_levels(context, port_id, host):
    if host:
        port_obj.PortBindingLevel.delete_objects(
            context, port_id=port_id, host=host)
        LOG.debug("For port %(port_id)s, host %(host)s, "
                  "cleared binding levels",
                  {'port_id': port_id,
                   'host': host})


def ensure_distributed_port_binding(context, port_id, host, router_id=None):
    with db_api.CONTEXT_READER.using(context):
        record = (context.session.query(models.DistributedPortBinding).
                  filter_by(port_id=port_id, host=host).first())
    if record:
        return record

    try:
        with db_api.CONTEXT_WRITER.using(context):
            record = models.DistributedPortBinding(
                port_id=port_id,
                host=host,
                router_id=router_id,
                vif_type=portbindings.VIF_TYPE_UNBOUND,
                vnic_type=portbindings.VNIC_NORMAL,
                status=n_const.PORT_STATUS_DOWN)
            context.session.add(record)
            return record
    except db_exc.DBDuplicateEntry:
        LOG.debug("Distributed Port %s already bound", port_id)
        with db_api.CONTEXT_READER.using(context):
            return (context.session.query(models.DistributedPortBinding).
                    filter_by(port_id=port_id, host=host).one())


def delete_distributed_port_binding_if_stale(context, binding):
    if not binding.router_id and binding.status == n_const.PORT_STATUS_DOWN:
        with db_api.CONTEXT_WRITER.using(context):
            LOG.debug("Distributed port: Deleting binding %s", binding)
            context.session.delete(binding)
            for bindlv in (context.session.query(models.PortBindingLevel).
                           filter_by(port_id=binding.port_id,
                                     host=binding.host)):
                context.session.delete(bindlv)
            LOG.debug("For port %(port_id)s, host %(host)s, "
                      "cleared binding levels",
                      {'port_id': binding.port_id,
                       'host': binding.host})


def get_port(context, port_id):
    """Get port record for update within transaction."""

    with db_api.CONTEXT_READER.using(context):
        try:
            # Set enable_eagerloads to True, so that lazy load can be
            # proceed later.
            record = (context.session.query(models_v2.Port).
                      enable_eagerloads(True).
                      filter(models_v2.Port.id.startswith(port_id)).
                      one())
            return record
        except exc.NoResultFound:
            return
        except exc.MultipleResultsFound:
            LOG.error("Multiple ports have port_id starting with %s",
                      port_id)
            return


def get_ports_and_sgs(context, port_ids):
    """Get ports from database with security group info."""

    # break large queries into smaller parts
    if len(port_ids) > MAX_PORTS_PER_QUERY:
        LOG.debug("Number of ports %(pcount)s exceeds the maximum per "
                  "query %(maxp)s. Partitioning queries.",
                  {'pcount': len(port_ids), 'maxp': MAX_PORTS_PER_QUERY})
        return (get_ports_and_sgs(context, port_ids[:MAX_PORTS_PER_QUERY]) +
                get_ports_and_sgs(context, port_ids[MAX_PORTS_PER_QUERY:]))

    LOG.debug("get_ports_and_sgs() called for port_ids %s", port_ids)

    if not port_ids:
        # if port_ids is empty, avoid querying to DB to ask it for nothing
        return []
    ports_to_sg_ids = get_sg_ids_grouped_by_port(context, port_ids)
    return [make_port_dict_with_security_groups(port, sec_groups)
            for port, sec_groups in ports_to_sg_ids.items()]


def get_sg_ids_grouped_by_port(context, port_ids):
    sg_ids_grouped_by_port = {}
    sg_binding_port = sg_models.SecurityGroupPortBinding.port_id

    with db_api.CONTEXT_READER.using(context):
        # partial UUIDs must be individually matched with startswith.
        # full UUIDs may be matched directly in an IN statement
        partial_uuids = set(port_id for port_id in port_ids
                            if not uuidutils.is_uuid_like(port_id))
        full_uuids = set(port_ids) - partial_uuids
        or_criteria = [models_v2.Port.id.startswith(port_id)
                       for port_id in partial_uuids]
        if full_uuids:
            or_criteria.append(models_v2.Port.id.in_(full_uuids))

        query = context.session.query(
            models_v2.Port,
            sg_models.SecurityGroupPortBinding.security_group_id)
        query = query.outerjoin(sg_models.SecurityGroupPortBinding,
                                models_v2.Port.id == sg_binding_port)
        query = query.filter(or_(*or_criteria))

        for port, sg_id in query:
            if port not in sg_ids_grouped_by_port:
                sg_ids_grouped_by_port[port] = []
            if sg_id:
                sg_ids_grouped_by_port[port].append(sg_id)
    return sg_ids_grouped_by_port


def make_port_dict_with_security_groups(port, sec_groups):
    plugin = directory.get_plugin()
    port_dict = plugin._make_port_dict(port)
    port_dict['security_groups'] = sec_groups
    port_dict['security_group_rules'] = []
    port_dict['security_group_source_groups'] = []
    port_dict['security_group_remote_address_groups'] = []
    port_dict['fixed_ips'] = [ip['ip_address']
                              for ip in port['fixed_ips']]
    return port_dict


def get_port_binding_host(context, port_id):
    try:
        with db_api.CONTEXT_READER.using(context):
            query = (context.session.query(models.PortBinding.host).
                     filter(models.PortBinding.port_id.startswith(port_id)))
            query = query.filter(
                models.PortBinding.status == n_const.ACTIVE).one()
    except exc.NoResultFound:
        LOG.debug("No active binding found for port %(port_id)s",
                  {'port_id': port_id})
        return
    except exc.MultipleResultsFound:
        LOG.error("Multiple ports have port_id starting with %s",
                  port_id)
        return
    return query.host


@db_api.CONTEXT_READER
def generate_distributed_port_status(context, port_id):
    # an OR'ed value of status assigned to parent port from the
    # distributedportbinding bucket
    query = context.session.query(models.DistributedPortBinding.status)
    final_status = n_const.PORT_STATUS_BUILD
    for bind in query.filter(models.DistributedPortBinding.port_id == port_id):
        if bind.status == n_const.PORT_STATUS_ACTIVE:
            return bind.status
        elif bind.status == n_const.PORT_STATUS_DOWN:
            final_status = bind.status
    return final_status


def get_distributed_port_binding_by_host(context, port_id, host):
    with db_api.CONTEXT_READER.using(context):
        binding = (
            context.session.query(models.DistributedPortBinding).
            filter(models.DistributedPortBinding.port_id.startswith(port_id),
                   models.DistributedPortBinding.host == host).first())
    if not binding:
        LOG.debug("No binding for distributed port %(port_id)s with host "
                  "%(host)s", {'port_id': port_id, 'host': host})
    return binding


def update_distributed_port_binding_by_host(context, port_id, host, router_id):
    with db_api.CONTEXT_WRITER.using(context):
        bindings = (
            context.session.query(models.DistributedPortBinding).
            filter(models.DistributedPortBinding.port_id.startswith(port_id),
                   models.DistributedPortBinding.host == host).all())
        for binding in bindings or []:
            binding['router_id'] = router_id or None
            binding.update(binding)


def get_distributed_port_bindings(context, port_id):
    with db_api.CONTEXT_READER.using(context):
        bindings = (context.session.query(models.DistributedPortBinding).
                    filter(models.DistributedPortBinding.port_id.startswith(
                        port_id)).all())
    if not bindings:
        LOG.debug("No bindings for distributed port %s", port_id)
    return bindings


@db_api.CONTEXT_READER
def partial_port_ids_to_full_ids(context, partial_ids):
    """Takes a list of the start of port IDs and returns full IDs.

    Returns dictionary of partial IDs to full IDs if a single match
    is found.
    """
    result = {}
    to_full_query = (context.session.query(models_v2.Port.id).
                     filter(or_(*[models_v2.Port.id.startswith(p)
                                  for p in partial_ids])))
    candidates = [match[0] for match in to_full_query]
    for partial_id in partial_ids:
        matching = [c for c in candidates if c.startswith(partial_id)]
        if len(matching) == 1:
            result[partial_id] = matching[0]
            continue
        if len(matching) < 1:
            LOG.info("No ports have port_id starting with %s", partial_id)
        elif len(matching) > 1:
            LOG.error("Multiple ports have port_id starting with %s",
                      partial_id)
    return result


@db_api.CONTEXT_READER
def get_port_db_objects(context, port_ids):
    """Takes a list of port_ids and returns matching port db objects.

    return format is a dictionary keyed by passed in IDs with db objects
    for values or None if the port was not present.
    """
    port_qry = (context.session.query(models_v2.Port).
                filter(models_v2.Port.id.in_(port_ids)))
    result = {p: None for p in port_ids}
    for port in port_qry:
        result[port.id] = port
    return result


@db_api.CONTEXT_READER
def is_dhcp_active_on_any_subnet(context, subnet_ids):
    if not subnet_ids:
        return False
    return bool(context.session.query(models_v2.Subnet.id).
                enable_eagerloads(False).filter_by(enable_dhcp=True).
                filter(models_v2.Subnet.id.in_(subnet_ids)).count())


def _prevent_segment_delete_with_port_bound(resource, event, trigger,
                                            payload=None):
    """Raise exception if there are any ports bound with segment_id."""
    if payload.metadata.get(seg_db.FOR_NET_DELETE):
        # don't check for network deletes
        return

    auto_delete_port_ids, proper_port_count = port_obj.Port.\
        get_auto_deletable_port_ids_and_proper_port_count_by_segment(
            payload.context, segment_id=payload.resource_id)

    if proper_port_count:
        reason = (_("The segment is still bound with %s port(s)") %
                  (proper_port_count + len(auto_delete_port_ids)))
        raise seg_exc.SegmentInUse(segment_id=payload.resource_id,
                                   reason=reason)

    if auto_delete_port_ids:
        LOG.debug("Auto-deleting dhcp port(s) on segment %s: %s",
                  payload.resource_id, ", ".join(auto_delete_port_ids))
        plugin = directory.get_plugin()
    for port_id in auto_delete_port_ids:
        try:
            plugin.delete_port(payload.context.elevated(), port_id)
        except nlib_exc.PortNotFound:
            # Don't raise if something else concurrently deleted the port
            LOG.debug("Ignoring PortNotFound when deleting port '%s'. "
                      "The port has already been deleted.", port_id)


def subscribe():
    registry.subscribe(_prevent_segment_delete_with_port_bound,
                       resources.SEGMENT,
                       events.BEFORE_DELETE)


subscribe()



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import functools
from random import randint
import re
import threading
import uuid

import netaddr
from neutron_lib import constants as n_const
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log
from oslo_utils import netutils
from ovsdbapp.backend.ovs_idl import event as row_event
from ovsdbapp.backend.ovs_idl import vlog

from neutron.agent.linux import external_process
from neutron.agent.linux import ip_lib
from neutron.agent.linux import iptables_manager
from neutron.agent.ovn.metadata import driver as metadata_driver
from neutron.agent.ovn.metadata import ovsdb
from neutron.agent.ovn.metadata import server as metadata_server
from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import utils as ovn_utils
from neutron.common import utils
from neutron.conf.agent.database import agents_db
from neutron.conf.plugins.ml2.drivers.ovn import ovn_conf as config


LOG = log.getLogger(__name__)
agents_db.register_db_agents_opts()
_SYNC_STATE_LOCK = lockutils.ReaderWriterLock()
CHASSIS_METADATA_LOCK = 'chassis_metadata_lock'

NS_PREFIX = ovn_const.OVN_METADATA_PREFIX
MAC_PATTERN = re.compile(r'([0-9A-F]{2}[:-]){5}([0-9A-F]{2})', re.I)
OVN_VIF_PORT_TYPES = (
    "", ovn_const.LSP_TYPE_EXTERNAL, ovn_const.LSP_TYPE_LOCALPORT)

MetadataPortInfo = collections.namedtuple('MetadataPortInfo', ['mac',
                                                               'ip_addresses',
                                                               'logical_port'])

OVN_METADATA_UUID_NAMESPACE = uuid.UUID('d34bf9f6-da32-4871-9af8-15a4626b41ab')


def _sync_lock(f):
    """Decorator to block all operations for a global sync call."""
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
        with _SYNC_STATE_LOCK.write_lock():
            return f(*args, **kwargs)
    return wrapped


def _match_only_if_additional_chassis_is_supported(f):
    @functools.wraps(f)
    def wrapped(self, row, old):
        if not ovn_utils.is_additional_chassis_supported(self.agent.sb_idl):
            return False
        return f(self, row, old)
    return wrapped


class ConfigException(Exception):
    """Misconfiguration of the agent

    This exception is raised when agent detects its wrong configuration.
    Typically agent should resync when this is raised.
    """


class PortBindingEvent(row_event.RowEvent):
    def __init__(self, metadata_agent):
        self.agent = metadata_agent
        table = 'Port_Binding'
        super().__init__((self.__class__.EVENT,), table, None)
        self.event_name = self.__class__.__name__
        self._log_msg = (
            "PortBindingEvent matched for logical port %s and network %s")

    def log_row(self, row):
        net_name = ovn_utils.get_network_name_from_datapath(
            row.datapath)
        LOG.info(self._log_msg, row.logical_port, net_name)

    def match_fn(self, event, row, old):
        return row.type in OVN_VIF_PORT_TYPES

    def run(self, event, row, old):
        # Check if the port has been bound/unbound to our chassis and update
        # the metadata namespace accordingly.
        resync = False

        with _SYNC_STATE_LOCK.read_lock():
            self.log_row(row)
            try:
                self.agent.provision_datapath(row.datapath)
            except ConfigException:
                # We're now in the reader lock mode, we need to exit the
                # context and then use writer lock
                resync = True
        if resync:
            self.agent.resync()


class PortBindingCreateWithChassis(PortBindingEvent):
    EVENT = PortBindingEvent.ROW_CREATE

    def match_fn(self, event, row, old):
        self._log_msg = "Port %s in datapath %s bound to our chassis on insert"
        if not (super().match_fn(event, row, old) and row.chassis):
            return False
        return row.chassis[0].name == self.agent.chassis


class PortBindingUpdatedEvent(PortBindingEvent):
    EVENT = PortBindingEvent.ROW_UPDATE

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._match_checks = [
            self._is_localport_ext_ids_update,
            self._is_new_chassis_set,
            self._is_chassis_removed,
            self._additional_chassis_added,
            self._additional_chassis_removed,
        ]

    def match_fn(self, event, row, old):
        if not super().match_fn(event, row, old):
            return False
        # if any of the check functions is true, the event should be triggered
        return any(check(row, old) for check in self._match_checks)

    def _is_localport_ext_ids_update(self, row, old):
        if row.type != ovn_const.LSP_TYPE_LOCALPORT:
            return False

        if not hasattr(old, 'external_ids'):
            return False

        device_id = row.external_ids.get(
            ovn_const.OVN_DEVID_EXT_ID_KEY, "")
        if not device_id.startswith(NS_PREFIX):
            return False

        new_cidrs = row.external_ids.get(
            ovn_const.OVN_CIDRS_EXT_ID_KEY, "")
        old_cidrs = old.external_ids.get(
            ovn_const.OVN_CIDRS_EXT_ID_KEY, "")
        # If old_cidrs is "", it is create event,
        # nothing needs to be done.
        # If old_cidrs equals new_cidrs, the ip does not change.
        if old_cidrs not in ("", new_cidrs):
            self._log_msg = (
                "Metadata Port %s in datapath %s updated")
            return True
        return False

    def _is_new_chassis_set(self, row, old):
        self._log_msg = "Port %s in datapath %s bound to our chassis"
        try:
            if ovn_utils.is_additional_chassis_supported(self.agent.sb_idl):
                try:
                    # If the additional chassis used to be in the old version
                    # the resources are already provisioned
                    if self.agent.chassis in {c.name for c in
                                              old.additional_chassis}:
                        return False
                except AttributeError:
                    pass
            return (row.chassis[0].name == self.agent.chassis and
                    not old.chassis)
        except (IndexError, AttributeError):
            return False

    def _is_chassis_removed(self, row, old):
        self._log_msg = "Port %s in datapath %s unbound from our chassis"
        try:
            return (old.chassis[0].name == self.agent.chassis and
                    not row.chassis)
        except (IndexError, AttributeError):
            return False

    @_match_only_if_additional_chassis_is_supported
    def _additional_chassis_added(self, row, old):
        # Additional chassis of the target node is set during an instance
        # live migration. We can provision resources early before the
        # instance lands on this chassis. After the VM finishes live
        # migration, it already has the resources provisioned therefore we
        # do not need to check when the chassis is moved from
        # the Additional_Chassis column to the Chassis column.
        additional_chassis = {ch for ch in row.additional_chassis
                              if ch.name == self.agent.chassis}
        self.log_msg = (
            "Live migrating port %s from network %s was added to this "
            "chassis. Provisioning resources early.")
        try:
            # Return True if the agent chassis was added to additional_chassis
            # column
            return bool(
                additional_chassis.difference(old.additional_chassis))
        except AttributeError:
            # If additional_chassis column was not changed then the old object
            # raises AttributeError when reading the column
            return False

    @_match_only_if_additional_chassis_is_supported
    def _additional_chassis_removed(self, row, old):
        # The method needs to check only for a case when agent chassis was set
        # in additional_chassis column, was removed but at the same time the
        # agent chassis was not set to chassis column. If the agent chassis is
        # set to chassis then it means live migration was successful and we do
        # not need to teardown the resources.
        try:
            old_a_chassis = {ch for ch in old.additional_chassis
                             if ch.name == self.agent.chassis}
        except AttributeError:
            # If additional chassis was not updated, the old object has no
            # additional_chassis attribute and raises an AttributeError
            return False

        # If was changed to the agent chassis then we do not need to teardown
        # the resources
        try:
            if (hasattr(old, 'chassis') and
                    row.chassis[0].name == self.agent.chassis):
                return False
        except IndexError:
            pass
        # We match the event only if the agent chassis was in the old
        # additional_chassis column and was removed
        return bool(old_a_chassis.difference(row.additional_chassis))


class PortBindingDeletedEvent(PortBindingEvent):
    EVENT = PortBindingEvent.ROW_DELETE

    def match_fn(self, event, row, old):
        if not super().match_fn(event, row, old):
            return False
        try:
            if row.chassis[0].name != self.agent.chassis:
                return False
        except (IndexError, AttributeError):
            return False
        if row.type != ovn_const.LSP_TYPE_EXTERNAL:
            LOG.warning(
                'Removing non-external type port %(port_id)s with '
                'type "%(type)s"',
                {"port_id": row.uuid, "type": row.type})
        self._log_msg = (
            "Port %s in datapath %s unbound from our chassis")
        return True


class ChassisPrivateCreateEvent(row_event.RowEvent):
    """Row create event - Chassis name == our_chassis.

    On connection, we get a dump of all chassis so if we catch a creation
    of our own chassis it has to be a reconnection. In this case, we need
    to do a full sync to make sure that we capture all changes while the
    connection to OVSDB was down.
    """
    def __init__(self, metadata_agent):
        self.agent = metadata_agent
        self.first_time = True
        events = (self.ROW_CREATE,)
        super(ChassisPrivateCreateEvent, self).__init__(
            events, 'Chassis_Private', (('name', '=', self.agent.chassis),))
        self.event_name = self.__class__.__name__

    def run(self, event, row, old):
        if self.first_time:
            self.first_time = False
        else:
            # NOTE(lucasagomes): Re-register the ovn metadata agent
            # with the local chassis in case its entry was re-created
            # (happens when restarting the ovn-controller)
            self.agent.register_metadata_agent()
            LOG.info("Connection to OVSDB established, doing a full sync")
            self.agent.sync()


class SbGlobalUpdateEvent(row_event.RowEvent):
    """Row update event on SB_Global table."""

    def __init__(self, metadata_agent):
        self.agent = metadata_agent
        table = 'SB_Global'
        events = (self.ROW_UPDATE,)
        super(SbGlobalUpdateEvent, self).__init__(events, table, None)
        self.event_name = self.__class__.__name__
        self.first_run = True

    def run(self, event, row, old):

        def _update_chassis(self, row):
            self.agent.sb_idl.db_set(
                'Chassis_Private', self.agent.chassis, ('external_ids', {
                    ovn_const.OVN_AGENT_METADATA_SB_CFG_KEY:
                        str(row.nb_cfg)})).execute()

        delay = 0
        if self.first_run:
            self.first_run = False
        else:
            # We occasionally see port binding failed errors due to
            # the ml2 driver refusing to bind the port to a dead agent.
            # if all agents heartbeat at the same time, they will all
            # cause a load spike on the server. To mitigate that we
            # need to spread out the load by introducing a random delay.
            # clamp the max delay between 3 and 10 seconds.
            max_delay = max(min(cfg.CONF.agent_down_time // 3, 10), 3)
            delay = randint(0, max_delay)

        LOG.debug("Delaying updating chassis table for %s seconds", delay)
        timer = threading.Timer(delay, _update_chassis, [self, row])
        timer.start()


class MetadataAgent(object):

    def __init__(self, conf):
        self.conf = conf
        vlog.use_python_logger(max_level=config.get_ovn_ovsdb_log_level())
        self._process_monitor = external_process.ProcessMonitor(
            config=self.conf,
            resource_type='metadata')
        self._sb_idl = None
        self._post_fork_event = threading.Event()
        # We'll restart all haproxy instances upon start so that they honor
        # any potential changes in their configuration.
        self.restarted_metadata_proxy_set = set()

    @property
    def sb_idl(self):
        if not self._sb_idl:
            self._post_fork_event.wait()
        return self._sb_idl

    @sb_idl.setter
    def sb_idl(self, val):
        self._sb_idl = val

    def _load_config(self):
        self.chassis = self._get_own_chassis_name()
        try:
            self.chassis_id = uuid.UUID(self.chassis)
        except ValueError:
            # OVS system-id could be a non UUID formatted string.
            self.chassis_id = uuid.uuid5(OVN_METADATA_UUID_NAMESPACE,
                                         self.chassis)
        self.ovn_bridge = self._get_ovn_bridge()
        LOG.debug("Loaded chassis name %s (UUID: %s) and ovn bridge %s.",
                  self.chassis, self.chassis_id, self.ovn_bridge)

    @_sync_lock
    def resync(self):
        """Resync the agent.

        Reload the configuration and sync the agent again.
        """
        self._load_config()
        self.sync()

    def start(self):
        # Open the connection to OVS database
        self.ovs_idl = ovsdb.MetadataAgentOvsIdl().start()
        self._load_config()

        tables = ('Encap', 'Port_Binding', 'Datapath_Binding', 'SB_Global',
                  'Chassis', 'Chassis_Private')
        events = (PortBindingUpdatedEvent(self),
                  PortBindingCreateWithChassis(self),
                  PortBindingDeletedEvent(self),
                  SbGlobalUpdateEvent(self),
                  ChassisPrivateCreateEvent(self),
                  )

        self._post_fork_event.clear()
        self.sb_idl = ovsdb.MetadataAgentOvnSbIdl(
            chassis=self.chassis, tables=tables, events=events).start()

        # Now IDL connections can be safely used.
        self._post_fork_event.set()

        # Launch the server that will act as a proxy between the VM's and Nova.
        self._proxy = metadata_server.UnixDomainMetadataProxy(
            self.conf, self.chassis, sb_idl=self.sb_idl)
        self._proxy.run()

        # Do the initial sync.
        # Provisioning handled by PortBindingCreateWithChassis
        self.sync(provision=False)

        # Register the agent with its corresponding Chassis
        self.register_metadata_agent()

        self._proxy.wait()

    @ovn_utils.retry()
    def register_metadata_agent(self):
        # NOTE(lucasagomes): db_add() will not overwrite the UUID if
        # it's already set.
        # Generate unique, but consistent metadata id for chassis name
        agent_id = uuid.uuid5(self.chassis_id, 'metadata_agent')
        ext_ids = {ovn_const.OVN_AGENT_METADATA_ID_KEY: str(agent_id)}
        self.sb_idl.db_add('Chassis_Private', self.chassis, 'external_ids',
                           ext_ids).execute(check_error=True)

    def _get_own_chassis_name(self):
        """Return the external_ids:system-id value of the Open_vSwitch table.

        As long as ovn-controller is running on this node, the key is
        guaranteed to exist and will include the chassis name.
        """
        ext_ids = self.ovs_idl.db_get(
            'Open_vSwitch', '.', 'external_ids').execute()
        return ext_ids['system-id']

    def _get_ovn_bridge(self):
        """Return the external_ids:ovn-bridge value of the Open_vSwitch table.

        This is the OVS bridge used to plug the metadata ports to.
        If the key doesn't exist, this method will return 'br-int' as default.
        """
        ext_ids = self.ovs_idl.db_get(
            'Open_vSwitch', '.', 'external_ids').execute()
        try:
            return ext_ids['ovn-bridge']
        except KeyError:
            LOG.warning("Can't read ovn-bridge external-id from OVSDB. Using "
                        "br-int instead.")
            return 'br-int'

    def get_networks_datapaths(self):
        """Return a set of datapath objects of the VIF ports on the current
        chassis.
        """
        ports = self.sb_idl.get_ports_on_chassis(
            self.chassis, include_additional_chassis=True)
        return set(p.datapath for p in self._vif_ports(ports))

    @_sync_lock
    def sync(self, provision=True):
        """Agent sync.

        This function will make sure that all networks with ports in our
        chassis are serving metadata. Also, it will tear down those namespaces
        which were serving metadata but are no longer needed.
        """

        # first, clean up namespaces that should no longer deploy
        system_namespaces = tuple(
            ns.decode('utf-8') if isinstance(ns, bytes) else ns
            for ns in ip_lib.list_network_namespaces())
        net_datapaths = self.get_networks_datapaths()
        metadata_namespaces = [
            self._get_namespace_name(
                ovn_utils.get_network_name_from_datapath(datapath))
            for datapath in net_datapaths
        ]
        unused_namespaces = [ns for ns in system_namespaces if
                             ns.startswith(NS_PREFIX) and
                             ns not in metadata_namespaces]
        for ns in unused_namespaces:
            try:
                self.teardown_datapath(self._get_datapath_name(ns))
            except Exception:
                LOG.exception('Error unable to destroy namespace: %s', ns)

        # resync all network namespaces based on the associated datapaths,
        # even those that are already running. This is to make sure
        # everything within each namespace is up to date.
        if provision:
            for datapath in net_datapaths:
                self.provision_datapath(datapath)

    @staticmethod
    def _get_veth_name(datapath):
        return ['{}{}{}'.format(n_const.TAP_DEVICE_PREFIX,
                                datapath[:10], i) for i in [0, 1]]

    @staticmethod
    def _get_datapath_name(namespace):
        return namespace[len(NS_PREFIX):]

    @staticmethod
    def _get_namespace_name(datapath):
        return NS_PREFIX + datapath

    def _vif_ports(self, ports):
        return (p for p in ports if p.type in OVN_VIF_PORT_TYPES)

    def teardown_datapath(self, net_name):
        """Unprovision this datapath to stop serving metadata.

        This function will shutdown metadata proxy if it's running and delete
        the VETH pair, the OVS port and the namespace.
        """
        namespace = self._get_namespace_name(net_name)
        ip = ip_lib.IPWrapper(namespace)
        # If the namespace doesn't exist, return
        if not ip.netns.exists(namespace):
            return

        LOG.info("Cleaning up %s namespace which is not needed anymore",
                 namespace)

        metadata_driver.MetadataDriver.destroy_monitored_metadata_proxy(
            self._process_monitor, net_name, self.conf, namespace)

        veth_name = self._get_veth_name(net_name)
        self.ovs_idl.del_port(veth_name[0]).execute()
        if ip_lib.device_exists(veth_name[0]):
            ip_lib.IPWrapper().del_veth(veth_name[0])

        ip.garbage_collect_namespace()

    def _ensure_datapath_checksum(self, namespace):
        """Ensure the correct checksum in the metadata packets in DPDK bridges

        (LP#1904871) In DPDK deployments (integration bridge datapath_type ==
        "netdev"), the checksum between the metadata namespace and OVS is not
        correctly populated.
        """
        if (self.ovs_idl.db_get(
                'Bridge', self.ovn_bridge, 'datapath_type').execute() !=
                ovn_const.CHASSIS_DATAPATH_NETDEV):
            return

        iptables_mgr = iptables_manager.IptablesManager(
            use_ipv6=netutils.is_ipv6_enabled(), nat=False,
            namespace=namespace, external_lock=False)
        rule = '-p tcp -m tcp -j CHECKSUM --checksum-fill'
        iptables_mgr.ipv4['mangle'].add_rule('POSTROUTING', rule, wrap=False)
        iptables_mgr.apply()

    def _get_port_ip4_ips(self, port):
        # Retrieve IPv4 addresses from the port mac column which is in form
        # ["<port_mac> <ip1> <ip2> ... <ipN>"]
        if not port.mac:
            LOG.warning("Port %s MAC column is empty, cannot retrieve IP "
                        "addresses", port.uuid)
            return []
        mac_field_attrs = port.mac[0].split()
        ips = mac_field_attrs[1:]
        if not ips:
            LOG.debug("Port %s IP addresses were not retrieved from the "
                      "Port_Binding MAC column %s", port.uuid, mac_field_attrs)
        return [ip for ip in ips if (
            utils.get_ip_version(ip) == n_const.IP_VERSION_4)]

    def _active_subnets_cidrs(self, datapath_ports_ips, metadata_port_cidrs):
        active_subnets_cidrs = set()
        # Prepopulate a dictionary where each metadata_port_cidr(string) maps
        # to its netaddr.IPNetwork object. This is so we dont have to
        # reconstruct IPNetwork objects repeatedly in the for loop
        metadata_cidrs_to_network_objects = {
            metadata_port_cidr: netaddr.IPNetwork(metadata_port_cidr)
            for metadata_port_cidr in metadata_port_cidrs if metadata_port_cidr
        }

        for datapath_port_ip in datapath_ports_ips:
            ip_obj = netaddr.IPAddress(datapath_port_ip)
            for metadata_cidr, metadata_cidr_obj in \
                    metadata_cidrs_to_network_objects.items():
                if ip_obj in metadata_cidr_obj:
                    active_subnets_cidrs.add(metadata_cidr)
                    break
        return active_subnets_cidrs

    def _process_cidrs(self, current_namespace_cidrs,
                       datapath_ports_ips, metadata_port_subnet_cidrs, lla):
        active_subnets_cidrs = self._active_subnets_cidrs(
            datapath_ports_ips, metadata_port_subnet_cidrs)

        cidrs_to_add = active_subnets_cidrs - current_namespace_cidrs

        # Make sure that all addresses, including the LLA, are present
        for addr in (n_const.METADATA_CIDR, n_const.METADATA_V6_CIDR, lla):
            if addr not in current_namespace_cidrs:
                cidrs_to_add.add(addr)
            else:
                active_subnets_cidrs.add(addr)

        cidrs_to_delete = current_namespace_cidrs - active_subnets_cidrs

        return cidrs_to_add, cidrs_to_delete

    def _get_provision_params(self, datapath):
        """Performs datapath preprovision checks and returns paremeters
        needed to provision namespace.

        Function will confirm that:
        1. Datapath metadata port has valid MAC
        2. There are datapath port IPs

        If any of those rules are not valid the nemaspace for the
        provided datapath will be tore down.
        If successful, returns datapath's network name, ports IPs
        and meta port info
        """
        net_name = ovn_utils.get_network_name_from_datapath(datapath)
        datapath_uuid = str(datapath.uuid)

        metadata_port = self.sb_idl.get_metadata_port_network(datapath_uuid)
        # If there's no metadata port or it doesn't have a MAC address, then
        # tear the namespace down if needed.
        if not (metadata_port and metadata_port.mac):
            LOG.debug("There is no metadata port for network %s or it has no "
                      "MAC address configured, tearing the namespace "
                      "down if needed", net_name)
            self.teardown_datapath(net_name)
            return

        # First entry of the mac field must be the MAC address.
        match = MAC_PATTERN.match(metadata_port.mac[0].split(' ')[0])
        if not match:
            LOG.error("Metadata port for network %s doesn't have a MAC "
                      "address, tearing the namespace down if needed",
                      net_name)
            self.teardown_datapath(net_name)
            return

        mac = match.group()
        ip_addresses = set(
            metadata_port.external_ids[
                ovn_const.OVN_CIDRS_EXT_ID_KEY].split(' '))
        metadata_port_info = MetadataPortInfo(mac, ip_addresses,
                                              metadata_port.logical_port)

        chassis_ports = self.sb_idl.get_ports_on_chassis(
            self.chassis, include_additional_chassis=True)
        datapath_ports_ips = []
        for chassis_port in self._vif_ports(chassis_ports):
            if str(chassis_port.datapath.uuid) == datapath_uuid:
                datapath_ports_ips.extend(self._get_port_ip4_ips(chassis_port))

        if not datapath_ports_ips:
            LOG.debug("No valid VIF ports were found for network %s, "
                      "tearing the namespace down if needed", net_name)
            self.teardown_datapath(net_name)
            return

        return net_name, datapath_ports_ips, metadata_port_info

    def provision_datapath(self, datapath):
        """Provision the datapath so that it can serve metadata.

        This function will create the namespace and VETH pair if needed
        and assign the IP addresses to the interface corresponding to the
        metadata port of the network. It will also remove existing IP from
        the namespace if they are no longer needed.

        :param datapath: datapath object.
        :return: The metadata namespace name for the datapath or None
                 if namespace was not provisioned
        """

        provision_params = self._get_provision_params(datapath)
        if not provision_params:
            return
        net_name, datapath_ports_ips, metadata_port_info = provision_params

        LOG.info("Provisioning metadata for network %s", net_name)
        # Create the VETH pair if it's not created. Also the add_veth function
        # will create the namespace for us.
        namespace = self._get_namespace_name(net_name)
        veth_name = self._get_veth_name(net_name)

        ip1 = ip_lib.IPDevice(veth_name[0])
        if ip_lib.device_exists(veth_name[1], namespace):
            ip2 = ip_lib.IPDevice(veth_name[1], namespace)
        else:
            LOG.debug("Creating VETH %s in %s namespace", veth_name[1],
                      namespace)
            # Might happen that the end in the root namespace exists even
            # though the other end doesn't. Make sure we delete it first if
            # that's the case.
            if ip1.exists():
                ip1.link.delete()
            ip1, ip2 = ip_lib.IPWrapper().add_veth(
                veth_name[0], veth_name[1], namespace)

        # Configure the MAC address.
        ip2.link.set_address(metadata_port_info.mac)

        # Make sure both ends of the VETH are up
        ip1.link.set_up()
        ip2.link.set_up()

        cidrs_to_add, cidrs_to_delete = self._process_cidrs(
            {dev['cidr'] for dev in ip2.addr.list()},
            datapath_ports_ips,
            metadata_port_info.ip_addresses,
            ip_lib.get_ipv6_lladdr(metadata_port_info.mac)
        )

        # Delete any non active addresses from the network namespace
        if cidrs_to_delete:
            ip2.addr.delete_multiple(list(cidrs_to_delete))

        if cidrs_to_add:
            ip2.addr.add_multiple(list(cidrs_to_add))

        # Check that this port is not attached to any other OVS bridge. This
        # can happen when the OVN bridge changes (for example, during a
        # migration from ML2/OVS).
        ovs_bridges = set(self.ovs_idl.list_br().execute())
        try:
            ovs_bridges.remove(self.ovn_bridge)
        except KeyError:
            LOG.warning("Configured OVN bridge %s cannot be found in "
                        "the system. Resyncing the agent.", self.ovn_bridge)
            raise ConfigException()

        if ovs_bridges:
            with self.ovs_idl.transaction() as txn:
                for br in ovs_bridges:
                    txn.add(self.ovs_idl.del_port(veth_name[0], bridge=br,
                                                  if_exists=True))

        # Configure the OVS port and add external_ids:iface-id so that it
        # can be tracked by OVN.
        self.ovs_idl.add_port(self.ovn_bridge,
                              veth_name[0]).execute()
        self.ovs_idl.db_set(
            'Interface', veth_name[0],
            ('external_ids', {'iface-id':
                              metadata_port_info.logical_port})).execute()

        # Ensure the correct checksum in the metadata traffic.
        self._ensure_datapath_checksum(namespace)

        if net_name not in self.restarted_metadata_proxy_set:
            metadata_driver.MetadataDriver.destroy_monitored_metadata_proxy(
                self._process_monitor, net_name, self.conf, namespace)
            self.restarted_metadata_proxy_set.add(net_name)

        # Spawn metadata proxy if it's not already running.
        metadata_driver.MetadataDriver.spawn_monitored_metadata_proxy(
            self._process_monitor, namespace, n_const.METADATA_PORT,
            self.conf, bind_address=n_const.METADATA_V4_IP,
            network_id=net_name, bind_address_v6=n_const.METADATA_V6_IP,
            bind_interface=veth_name[1])



# Copyright 2017 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_db.sqlalchemy import utils as db_utils

from neutron.db.migration.alembic_migrations.versions.yoga.expand import \
    ba859d649675_add_indexes_to_rbacs as _migration
from neutron.tests.functional.db import test_migrations


class TestAddIndexesToRbacsMixin(object):
    """Validates binding_index for NetworkDhcpAgentBinding migration."""

    @staticmethod
    def get_index(table_indexes, column):
        for index in table_indexes:
            if [column] == index['column_names']:
                return True
        return False

    def _pre_upgrade_ba859d649675(self, engine):
        for table in _migration.OBJECTS:
            table_indexes = db_utils.get_indexes(engine, table + 'rbacs')
            for column in _migration.COLUMNS:
                self.assertFalse(self.get_index(table_indexes, column))

    def _check_ba859d649675(self, engine, data):
        for table in _migration.OBJECTS:
            table_indexes = db_utils.get_indexes(engine, table + 'rbacs')
            for column in _migration.COLUMNS:
                self.assertTrue(self.get_index(table_indexes, column))


class TestAddIndexesToRbacsMySQL(
        TestAddIndexesToRbacsMixin,
        test_migrations.TestWalkMigrationsMySQL):
    pass


class TestAddIndexesToRbacsPostgreSQL(
        TestAddIndexesToRbacsMixin,
        test_migrations.TestWalkMigrationsPostgreSQL):
    pass



#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

from unittest import mock

from neutron.agent.linux import ipset_manager
from neutron.tests import base

TEST_SET_ID = 'fake_sgid'
ETHERTYPE = 'IPv4'
TEST_SET_NAME = ipset_manager.IpsetManager.get_name(TEST_SET_ID, ETHERTYPE)
TEST_SET_NAME_NEW = TEST_SET_NAME + ipset_manager.SWAP_SUFFIX
FAKE_IPS = [('10.0.0.1', 'fa:16:3e:aa:bb:c1'),
            ('10.0.0.2', 'fa:16:3e:aa:bb:c2'),
            ('10.0.0.3', 'fa:16:3e:aa:bb:c3'),
            ('10.0.0.4', 'fa:16:3e:aa:bb:c4'),
            ('10.0.0.5', 'fa:16:3e:aa:bb:c5'),
            ('10.0.0.6', 'fa:16:3e:aa:bb:c6')]


class BaseIpsetManagerTest(base.BaseTestCase):
    def setUp(self):
        super(BaseIpsetManagerTest, self).setUp()
        self.ipset = ipset_manager.IpsetManager()
        self.execute = mock.patch.object(self.ipset, "execute").start()
        self.expected_calls = []
        self.expect_create()
        self.force_sorted_get_set_ips()

    def force_sorted_get_set_ips(self):
        """Force sorted responses by self.ipset._get_new/deleted_set_ips.

        _get_new/deleted_set_ips use internally sets and return randomly
        ordered responses. This method ensures sorted responses from them
        in order to guarantee call order in self.ipset.set_members.
        """
        original_get_new_set_ips = self.ipset._get_new_set_ips
        original_get_deleted_set_ips = self.ipset._get_deleted_set_ips

        def sorted_get_new_set_ips(set_name, expected_ips):
            unsorted = original_get_new_set_ips(set_name, expected_ips)
            return sorted(unsorted)

        def sorted_get_deleted_set_ips(set_name, expected_ips):
            unsorted = original_get_deleted_set_ips(set_name, expected_ips)
            return sorted(unsorted)

        mock.patch.object(self.ipset, '_get_new_set_ips',
                          side_effect=sorted_get_new_set_ips).start()
        mock.patch.object(self.ipset, '_get_deleted_set_ips',
                          side_effect=sorted_get_deleted_set_ips).start()

    def verify_mock_calls(self):
        self.execute.assert_has_calls(self.expected_calls, any_order=False)

    def expect_set(self, addresses):
        temp_input = ['create %s hash:net family inet' % TEST_SET_NAME_NEW]
        temp_input.extend('add %s %s' % (TEST_SET_NAME_NEW, ip)
                          for ip in self.ipset._sanitize_addresses(addresses))
        input = '\n'.join(temp_input)
        self.expected_calls.extend([
            mock.call(['ipset', 'restore', '-exist'],
                      process_input=input, run_as_root=True,
                      check_exit_code=True, privsep_exec=True),
            mock.call(['ipset', 'swap', TEST_SET_NAME_NEW, TEST_SET_NAME],
                      process_input=None, run_as_root=True,
                      check_exit_code=True, privsep_exec=True),
            mock.call(['ipset', 'destroy', TEST_SET_NAME_NEW],
                      process_input=None, run_as_root=True,
                      check_exit_code=False, privsep_exec=True)])

    def expect_add(self, addresses):
        self.expected_calls.extend(
            mock.call(['ipset', 'add', '-exist', TEST_SET_NAME, ip],
                      process_input=None, run_as_root=True,
                      check_exit_code=True, privsep_exec=True)
            for ip in self.ipset._sanitize_addresses(addresses))

    def expect_del(self, addresses):

        self.expected_calls.extend(
            mock.call(['ipset', 'del', TEST_SET_NAME, ip],
                      process_input=None, run_as_root=True,
                      check_exit_code=False, privsep_exec=True)
            for ip in self.ipset._sanitize_addresses(addresses))

    def expect_create(self):
        self.expected_calls.append(
            mock.call(['ipset', 'create', '-exist', TEST_SET_NAME,
                       'hash:net', 'family', 'inet'],
                      process_input=None, run_as_root=True,
                      check_exit_code=True, privsep_exec=True))

    def expect_destroy(self):
        self.expected_calls.append(
            mock.call(['ipset', 'destroy', TEST_SET_NAME],
                      process_input=None, run_as_root=True,
                      check_exit_code=False, privsep_exec=True))

    def add_first_ip(self):
        self.expect_set([FAKE_IPS[0]])
        self.ipset.set_members(TEST_SET_ID, ETHERTYPE, [FAKE_IPS[0]])

    def add_all_ips(self):
        self.expect_set(FAKE_IPS)
        self.ipset.set_members(TEST_SET_ID, ETHERTYPE, FAKE_IPS)


class IpsetManagerTestCase(BaseIpsetManagerTest):

    def test_set_name_exists(self):
        self.add_first_ip()
        self.assertTrue(self.ipset.set_name_exists('N' + ETHERTYPE +
                                                   TEST_SET_ID))

    def test_set_members_with_first_add_member(self):
        self.add_first_ip()
        self.verify_mock_calls()

    def test_set_members_adding_less_than_5(self):
        self.add_first_ip()
        self.expect_add(FAKE_IPS[1:5])
        self.ipset.set_members(TEST_SET_ID, ETHERTYPE, FAKE_IPS[0:5])
        self.verify_mock_calls()

    def test_set_members_deleting_less_than_5(self):
        self.add_all_ips()
        self.expect_del(FAKE_IPS[3:])
        self.ipset.set_members(TEST_SET_ID, ETHERTYPE, FAKE_IPS[0:3])
        self.verify_mock_calls()

    def test_set_members_adding_more_than_5(self):
        self.add_first_ip()
        self.expect_set(FAKE_IPS)
        self.ipset.set_members(TEST_SET_ID, ETHERTYPE, FAKE_IPS)
        self.verify_mock_calls()

    def test_set_members_adding_all_zero_ipv4(self):
        self.expect_set([('0.0.0.0/0', 'fa:16:3e:aa:bb:c1'), ])
        self.ipset.set_members(TEST_SET_ID, ETHERTYPE,
                               [('0.0.0.0/0', 'fa:16:3e:aa:bb:c1'), ])
        self.verify_mock_calls()

    def test_set_members_adding_all_zero_ipv6(self):
        self.expect_set([('::/0', 'fa:16:3e:aa:bb:c1'), ])
        self.ipset.set_members(TEST_SET_ID, ETHERTYPE,
                               [('::/0', 'fa:16:3e:aa:bb:c1'), ])
        self.verify_mock_calls()

    def test_destroy(self):
        self.add_first_ip()
        self.expect_destroy()
        self.ipset.destroy(TEST_SET_ID, ETHERTYPE)
        self.verify_mock_calls()



# Copyright 2022 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from pyroute2 import netlink
from pyroute2.netlink.rtnl import ifinfmsg

from neutron.privileged.agent import linux as priv_linux
from neutron.tests import base


class MakeSerializableTestCase(base.BaseTestCase):

    NLA_DATA1 = ifinfmsg.ifinfbase.state(data=b'54321')
    NLA_DATA2 = ifinfmsg.ifinfbase.state(data=b'abcdef')
    INPUT_1 = {'key1': 'value1', b'key2': b'value2', 'key3': ('a', 2),
               'key4': [1, 2, 'c'],
               b'key5': netlink.nla_slot('nla_name1', NLA_DATA1),
               'key6': netlink.nla_slot(b'nla_name2', NLA_DATA2)}
    OUTPUT_1 = {'key1': 'value1', 'key2': 'value2', 'key3': ('a', 2),
                'key4': [1, 2, 'c'],
                'key5': ['nla_name1', '54321'],
                'key6': ['nla_name2', 'abcdef']}

    def test_make_serializable(self):
        self.assertEqual(self.OUTPUT_1,
                         priv_linux.make_serializable(self.INPUT_1))



# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from neutron_lib import context as nctx
from neutron_lib.db import api as db_api
from neutron_lib.plugins import constants
from neutron_lib.plugins import directory
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_utils import uuidutils
from sqlalchemy.orm import session as se
from webob import exc

from neutron.db import models_v2
from neutron.objects import ports as port_obj
from neutron.tests.unit.plugins.ml2 import test_plugin


class TestRevisionPlugin(test_plugin.Ml2PluginV2TestCase):

    l3_plugin = ('neutron.tests.unit.extensions.test_extraroute.'
                 'TestExtraRouteL3NatServicePlugin')

    _extension_drivers = ['qos']

    def get_additional_service_plugins(self):
        p = super(TestRevisionPlugin, self).get_additional_service_plugins()
        p.update({'revision_plugin_name': 'revisions',
                  'qos_plugin_name': 'qos',
                  'tag_name': 'tag'})
        return p

    def setUp(self):
        cfg.CONF.set_override('extension_drivers',
                              self._extension_drivers,
                              group='ml2')
        super(TestRevisionPlugin, self).setUp()
        self.cp = directory.get_plugin()
        self.l3p = directory.get_plugin(constants.L3)
        self._ctx = nctx.get_admin_context()
        self._tenant_id = uuidutils.generate_uuid()

    @property
    def ctx(self):
        # TODO(kevinbenton): return ctx without expire_all after switch to
        # enginefacade complete. We expire_all here because the switch to
        # the new engine facade is resulting in changes being spread over
        # other sessions so we can end up getting stale reads in the parent
        # session if objects remain in the identity map.
        if not db_api.is_session_active(self._ctx.session):
            self._ctx.session.expire_all()
        return self._ctx

    def test_handle_expired_object(self):
        rp = directory.get_plugin('revision_plugin')
        with self.port():
            with db_api.CONTEXT_WRITER.using(self.ctx):
                ipal_objs = port_obj.IPAllocation.get_objects(self.ctx)
                if not ipal_objs:
                    raise Exception("No IP allocations available.")
                ipal_obj = ipal_objs[0]
                # load port into our session
                port = self.ctx.session.query(models_v2.Port).one()
                # simulate concurrent delete in another session
                other_ctx = nctx.get_admin_context()
                other_ctx.session.delete(
                    other_ctx.session.query(models_v2.Port).first()
                )
                other_ctx.session.flush()

                # ensure no attribute lookups are attempted on an
                # object deleted from the session when doing related
                # bumps
                self.ctx.session.expire(port)

                collected = rp._collect_related_tobump(
                    self.ctx.session, [ipal_obj], set())
                rp._bump_obj_revisions(
                    self.ctx.session, collected, version_check=False)

    def test_shared_network_create(self):
        # this test intends to run db_base_plugin_v2 -> create_network_db,
        # which in turn creates a Network and then a NetworkRBAC object.
        # An issue was observed with the revision_plugin which would interfere
        # with the flush process that occurs with these two connected objects,
        # creating two copies of the Network object in the Session and putting
        # it into an invalid state.
        with self.network(shared=True, as_admin=True):
            pass

    def test_port_name_update_revises(self):
        with self.port() as port:
            rev = port['port']['revision_number']
            new = {'port': {'name': 'seaweed'}}
            response = self._update('ports', port['port']['id'], new)
            new_rev = response['port']['revision_number']
            self.assertGreater(new_rev, rev)

    def test_constrained_port_update(self):
        with self.port() as port:
            rev = port['port']['revision_number']
            new = {'port': {'name': 'nigiri'}}
            for val in (rev - 1, rev + 1):
                # make sure off-by ones are rejected
                self._update('ports', port['port']['id'], new,
                             headers={'If-Match': 'revision_number=%s' % val},
                             expected_code=exc.HTTPPreconditionFailed.code)
            after_attempt = self._show('ports', port['port']['id'])
            self.assertEqual(rev, after_attempt['port']['revision_number'])
            self.assertEqual(port['port']['name'],
                             after_attempt['port']['name'])
            # correct revision should work
            self._update('ports', port['port']['id'], new,
                         headers={'If-Match': 'revision_number=%s' % rev})

    def test_constrained_port_delete(self):
        with self.port() as port:
            rev = port['port']['revision_number']
            for val in (rev - 1, rev + 1):
                # make sure off-by ones are rejected
                self._delete('ports', port['port']['id'],
                             headers={'If-Match': 'revision_number=%s' % val},
                             expected_code=exc.HTTPPreconditionFailed.code)
            # correct revision should work
            self._delete('ports', port['port']['id'],
                         headers={'If-Match': 'revision_number=%s' % rev})

    def test_constrained_port_update_handles_db_retries(self):
        # here we ensure all of the constraint handling logic persists
        # on retriable failures to commit caused by races with another
        # update
        with self.port() as port:
            rev = port['port']['revision_number']
            new = {'port': {'name': 'nigiri'}}

            def concurrent_increment(s):
                db_api.sqla_remove(se.Session, 'before_commit',
                                   concurrent_increment)
                # slip in a concurrent update that will bump the revision
                plugin = directory.get_plugin()
                plugin.update_port(nctx.get_admin_context(),
                                   port['port']['id'], new)
                raise db_exc.DBDeadlock()
            db_api.sqla_listen(se.Session, 'before_commit',
                               concurrent_increment)

            # Despite the revision number is bumped twice during the session
            # transaction, the revision number is tested only once the first
            # time the revision number service is executed for this session and
            # object.
            self._update('ports', port['port']['id'], new,
                         headers={'If-Match': 'revision_number=%s' % rev},
                         expected_code=exc.HTTPOk.code)
            self._update('ports', port['port']['id'], new,
                         headers={'If-Match': 'revision_number=%s' %
                                              str(int(rev) + 2)},
                         expected_code=exc.HTTPOk.code)
            self._update('ports', port['port']['id'], new,
                         headers={'If-Match': 'revision_number=1'},
                         expected_code=exc.HTTPPreconditionFailed.code)

    def test_port_ip_update_revises(self):
        with self.subnet() as subnet, self.port(subnet=subnet) as port:
            rev = port['port']['revision_number']
            new = {'port': {'fixed_ips': port['port']['fixed_ips']}}
            # ensure adding an IP allocation updates the port
            free_ip = self._find_ip_address(subnet['subnet'])
            new['port']['fixed_ips'].append({'ip_address': free_ip})
            response = self._update('ports', port['port']['id'], new)
            self.assertEqual(2, len(response['port']['fixed_ips']))
            new_rev = response['port']['revision_number']
            self.assertGreater(new_rev, rev)
            # ensure deleting an IP allocation updates the port
            rev = new_rev
            new['port']['fixed_ips'].pop()
            response = self._update('ports', port['port']['id'], new)
            self.assertEqual(1, len(response['port']['fixed_ips']))
            new_rev = response['port']['revision_number']
            self.assertGreater(new_rev, rev)

    def test_network_description_bumps_revision(self):
        with self.network() as net:
            rev = net['network']['revision_number']
            data = {'network': {'description': 'Test Description'}}
            response = self._update('networks', net['network']['id'], data)
            new_rev = response['network']['revision_number']
            self.assertEqual(rev + 1, new_rev)

    def test_subnet_description_bumps_revision(self):
        with self.subnet() as subnet:
            rev = subnet['subnet']['revision_number']
            data = {'subnet': {'description': 'Test Description'}}
            response = self._update('subnets', subnet['subnet']['id'], data)
            new_rev = response['subnet']['revision_number']
            self.assertEqual(rev + 1, new_rev)

    def test_security_group_rule_ops_bump_security_group(self):
        s = {'security_group': {'tenant_id': 'some_tenant', 'name': '',
                                'description': 's'}}
        sg = self.cp.create_security_group(self.ctx, s)
        s['security_group']['name'] = 'hello'
        updated = self.cp.update_security_group(self.ctx, sg['id'], s)
        self.assertGreater(updated['revision_number'], sg['revision_number'])
        # ensure rule changes bump parent SG
        r = {'security_group_rule': {'tenant_id': 'some_tenant',
                                     'port_range_min': 80, 'protocol': 6,
                                     'port_range_max': 90,
                                     'remote_ip_prefix': '0.0.0.0/0',
                                     'ethertype': 'IPv4',
                                     'remote_group_id': None,
                                     'remote_address_group_id': None,
                                     'direction': 'ingress',
                                     'security_group_id': sg['id']}}
        rule = self.cp.create_security_group_rule(self.ctx, r)
        sg = updated
        updated = self.cp.get_security_group(self.ctx, sg['id'])
        self.assertGreater(updated['revision_number'], sg['revision_number'])
        self.cp.delete_security_group_rule(self.ctx, rule['id'])
        sg = updated
        updated = self.cp.get_security_group(self.ctx, sg['id'])
        self.assertGreater(updated['revision_number'], sg['revision_number'])

    def test_router_interface_ops_bump_router(self):
        r = {'router': {'name': 'myrouter', 'tenant_id': 'some_tenant',
                        'admin_state_up': True}}
        router = self.l3p.create_router(self.ctx, r)
        r['router']['name'] = 'yourrouter'
        updated = self.l3p.update_router(self.ctx, router['id'], r)
        self.assertGreater(updated['revision_number'],
                           router['revision_number'])
        # add an intf and make sure it bumps rev
        with self.subnet(tenant_id='some_tenant', cidr='10.0.1.0/24') as s:
            interface_info = {'subnet_id': s['subnet']['id']}
        self.l3p.add_router_interface(self.ctx, router['id'],
                                      interface_info)
        router = updated
        updated = self.l3p.get_router(self.ctx, router['id'])
        self.assertGreater(updated['revision_number'],
                           router['revision_number'])
        # Add a route and make sure it bumps revision number
        router = updated
        body = {'router': {'routes': [{'destination': '192.168.2.0/24',
                                       'nexthop': '10.0.1.3'}]}}
        self.l3p.update_router(self.ctx, router['id'], body)
        updated = self.l3p.get_router(self.ctx, router['id'])
        self.assertGreater(updated['revision_number'],
                           router['revision_number'])
        router = updated
        body['router']['routes'] = []
        self.l3p.update_router(self.ctx, router['id'], body)
        updated = self.l3p.get_router(self.ctx, router['id'])
        self.assertGreater(updated['revision_number'],
                           router['revision_number'])
        self.l3p.remove_router_interface(self.ctx, router['id'],
                                         interface_info)
        router = updated
        updated = self.l3p.get_router(self.ctx, router['id'])
        self.assertGreater(updated['revision_number'],
                           router['revision_number'])

    def test_qos_policy_bump_port_revision(self):
        with self.port() as port:
            rev = port['port']['revision_number']
            qos_plugin = directory.get_plugin('QOS')
            qos_policy = {'policy': {'id': uuidutils.generate_uuid(),
                                     'name': "policy1",
                                     'project_id': uuidutils.generate_uuid()}}
            qos_obj = qos_plugin.create_policy(self.ctx, qos_policy)
            data = {'port': {'qos_policy_id': qos_obj['id']}}
            response = self._update('ports', port['port']['id'], data,
                                    as_admin=True)
            new_rev = response['port']['revision_number']
            self.assertGreater(new_rev, rev)

    def test_qos_policy_bump_network_revision(self):
        with self.network() as network:
            rev = network['network']['revision_number']
            qos_plugin = directory.get_plugin('QOS')
            qos_policy = {'policy': {'id': uuidutils.generate_uuid(),
                                     'name': "policy1",
                                     'project_id': uuidutils.generate_uuid()}}
            qos_obj = qos_plugin.create_policy(self.ctx, qos_policy)
            data = {'network': {'qos_policy_id': qos_obj['id']}}
            response = self._update('networks', network['network']['id'], data,
                                    as_admin=True)
            new_rev = response['network']['revision_number']
            self.assertGreater(new_rev, rev)

    def test_net_tag_bumps_net_revision(self):
        with self.network() as network:
            rev = network['network']['revision_number']
            tag_plugin = directory.get_plugin('TAG')
            tag_plugin.update_tag(self.ctx, 'networks',
                                  network['network']['id'], 'mytag')
            updated = directory.get_plugin().get_network(
                self.ctx, network['network']['id'])
            self.assertGreater(updated['revision_number'], rev)
            tag_plugin.delete_tag(self.ctx, 'networks',
                                  network['network']['id'], 'mytag')
            rev = updated['revision_number']
            updated = directory.get_plugin().get_network(
                self.ctx, network['network']['id'])
            self.assertGreater(updated['revision_number'], rev)



# Copyright (c) 2023 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovsdbapp.backend.ovs_idl import event as row_event

from neutron.agent.ovn.extensions import extension_manager as ext_mgr


class OVSInterfaceEvent(row_event.RowEvent):

    def __init__(self, ovn_agent):
        self.ovn_agent = ovn_agent
        events = (self.ROW_CREATE, self.ROW_UPDATE)
        table = 'Interface'
        super().__init__(events, table, None)

    def run(self, event, row, old):
        self.ovn_agent.test_ovs_idl.append(row.name)


class OVNSBChassisEvent(row_event.RowEvent):
    def __init__(self, ovn_agent):
        self.ovn_agent = ovn_agent
        events = (self.ROW_CREATE, self.ROW_UPDATE)
        table = 'Chassis'
        super().__init__(events, table, None)

    def run(self, event, row, old):
        self.ovn_agent.test_ovn_sb_idl.append(row.name)


class OVNNBLogicalSwitchEvent(row_event.RowEvent):
    def __init__(self, ovn_agent):
        self.ovn_agent = ovn_agent
        events = (self.ROW_CREATE, self.ROW_UPDATE)
        table = 'Logical_Switch'
        super().__init__(events, table, None)

    def run(self, event, row, old):
        self.ovn_agent.test_ovn_nb_idl.append(row.name)


class FakeOVNAgentExtension(ext_mgr.OVNAgentExtension):

    @property
    def name(self):
        return 'Fake OVN agent extension'

    @property
    def ovs_idl_events(self):
        return [OVSInterfaceEvent]

    @property
    def nb_idl_tables(self):
        return ['Logical_Switch']

    @property
    def nb_idl_events(self):
        return [OVNNBLogicalSwitchEvent]

    @property
    def sb_idl_tables(self):
        return ['Chassis', 'Chassis_Private']

    @property
    def sb_idl_events(self):
        return [OVNSBChassisEvent]



# Copyright 2014 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

# Initial operations for the port security extension


from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table(
        'networksecuritybindings',
        sa.Column('network_id', sa.String(length=36), nullable=False),
        sa.Column('port_security_enabled', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(['network_id'], ['networks.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('network_id'))

    op.create_table(
        'portsecuritybindings',
        sa.Column('port_id', sa.String(length=36), nullable=False),
        sa.Column('port_security_enabled', sa.Boolean(), nullable=False),
        sa.ForeignKeyConstraint(['port_id'], ['ports.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('port_id'))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import context as n_context
from neutron_lib.db import api as db_api
from neutron_lib import exceptions as n_exc
from neutron_lib.services.trunk import constants as trunk_consts
from oslo_config import cfg
from oslo_log import log

from neutron.common.ovn import constants as ovn_const
from neutron.common import utils as n_utils
from neutron.db import db_base_plugin_common
from neutron.db import ovn_revision_numbers_db as db_rev
from neutron.objects import ports as port_obj
from neutron.services.trunk.drivers import base as trunk_base
from neutron.services.trunk import exceptions as trunk_exc


SUPPORTED_INTERFACES = (
    portbindings.VIF_TYPE_OVS,
    portbindings.VIF_TYPE_VHOST_USER,
)

SUPPORTED_SEGMENTATION_TYPES = (
    trunk_consts.SEGMENTATION_TYPE_VLAN,
)

LOG = log.getLogger(__name__)


class OVNTrunkHandler(object):
    def __init__(self, plugin_driver):
        self.plugin_driver = plugin_driver

    def _set_sub_ports(self, parent_port, subports):
        txn = self.plugin_driver.nb_ovn.transaction
        context = n_context.get_admin_context()
        db_parent_port = port_obj.Port.get_object(context, id=parent_port)
        parent_port_status = db_parent_port.status
        for subport in subports:
            with db_api.CONTEXT_WRITER.using(context), (
                    txn(check_error=True)) as ovn_txn:
                port = self._set_binding_profile(context, subport, parent_port,
                                                 parent_port_status, ovn_txn)
            db_rev.bump_revision(context, port, ovn_const.TYPE_PORTS)

    def _unset_sub_ports(self, subports):
        txn = self.plugin_driver.nb_ovn.transaction
        context = n_context.get_admin_context()
        for subport in subports:
            with db_api.CONTEXT_WRITER.using(context), (
                    txn(check_error=True)) as ovn_txn:
                port = self._unset_binding_profile(context, subport, ovn_txn)
            db_rev.bump_revision(context, port, ovn_const.TYPE_PORTS)

    @db_base_plugin_common.convert_result_to_dict
    def _set_binding_profile(self, context, subport, parent_port,
                             parent_port_status, ovn_txn):
        LOG.debug("Setting parent %s for subport %s",
                  parent_port, subport.port_id)
        db_port = port_obj.Port.get_object(context, id=subport.port_id)
        if not db_port:
            LOG.debug("Port not found while trying to set "
                      "binding_profile: %s",
                      subport.port_id)
            return
        check_rev_cmd = self.plugin_driver.nb_ovn.check_revision_number(
            db_port.id, db_port, ovn_const.TYPE_PORTS)
        ovn_txn.add(check_rev_cmd)
        try:
            # NOTE(flaviof): We expect binding's host to be set. Otherwise,
            # sub-port will not transition from DOWN to ACTIVE.
            db_port.device_owner = trunk_consts.TRUNK_SUBPORT_OWNER
            # NOTE(ksambor):  When sub-port was created and event was process
            # without binding profile this port will end forever in DOWN
            # status so we need to switch it here to the parent port status
            db_port.status = parent_port_status
            for binding in db_port.bindings:
                binding.profile['parent_name'] = parent_port
                binding.profile['tag'] = subport.segmentation_id
                # host + port_id is primary key
                port_obj.PortBinding.update_object(
                    context,
                    {'profile': binding.profile,
                     'vif_type': portbindings.VIF_TYPE_OVS},
                    port_id=subport.port_id,
                    host=binding.host)
            db_port.update()
        except n_exc.ObjectNotFound:
            LOG.debug("Port not found while trying to set "
                      "binding_profile: %s", subport.port_id)
            return
        ext_ids = {ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY: db_port.device_owner}
        ovn_txn.add(self.plugin_driver.nb_ovn.set_lswitch_port(
            lport_name=subport.port_id,
            parent_name=parent_port,
            tag=subport.segmentation_id,
            external_ids_update=ext_ids,
        ))
        LOG.debug("Done setting parent %s for subport %s",
                  parent_port, subport.port_id)
        return db_port

    @db_base_plugin_common.convert_result_to_dict
    def _unset_binding_profile(self, context, subport, ovn_txn):
        LOG.debug("Unsetting parent for subport %s", subport.port_id)
        db_port = port_obj.Port.get_object(context, id=subport.port_id)
        if not db_port:
            LOG.debug("Port not found while trying to unset "
                      "binding_profile: %s",
                      subport.port_id)
            return
        check_rev_cmd = self.plugin_driver.nb_ovn.check_revision_number(
            db_port.id, db_port, ovn_const.TYPE_PORTS)
        ovn_txn.add(check_rev_cmd)
        try:
            db_port.device_owner = ''
            for binding in db_port.bindings:
                binding.profile.pop('tag', None)
                binding.profile.pop('parent_name', None)
                # host + port_id is primary key
                port_obj.PortBinding.update_object(
                    context,
                    {'profile': binding.profile,
                     'vif_type': portbindings.VIF_TYPE_UNBOUND},
                    port_id=subport.port_id,
                    host=binding.host)
                port_obj.PortBindingLevel.delete_objects(
                    context, port_id=subport.port_id, host=binding.host)
            db_port.update()
        except n_exc.ObjectNotFound:
            LOG.debug("Port not found while trying to unset "
                      "binding_profile: %s", subport.port_id)
            return
        ext_ids = {ovn_const.OVN_DEVICE_OWNER_EXT_ID_KEY: db_port.device_owner}
        ovn_txn.add(self.plugin_driver.nb_ovn.set_lswitch_port(
            lport_name=subport.port_id,
            parent_name=[],
            up=False,
            tag=[],
            external_ids_update=ext_ids,
        ))
        LOG.debug("Done unsetting parent for subport %s", subport.port_id)
        return db_port

    @staticmethod
    def _is_port_bound(port):
        return n_utils.is_port_bound(port, log_message=False)

    def trunk_created(self, resource, event, trunk_plugin, payload):
        trunk = payload.states[0]
        # Check if parent port is handled by OVN.
        if not self.plugin_driver.nb_ovn.lookup('Logical_Switch_Port',
                                                trunk.port_id, default=None):
            return
        if trunk.sub_ports:
            self._set_sub_ports(trunk.port_id, trunk.sub_ports)
        trunk.update(status=trunk_consts.TRUNK_ACTIVE_STATUS)

    def trunk_deleted(self, resource, event, trunk_plugin, payload):
        trunk = payload.states[0]
        if trunk.sub_ports:
            self._unset_sub_ports(trunk.sub_ports)

    def trunk_created_precommit(self, resource, event, trunk_plugin, payload):
        # payload.desired_state below is the trunk object
        parent_port = payload.desired_state.db_obj.port
        if self._is_port_bound(parent_port):
            raise trunk_exc.ParentPortInUse(port_id=parent_port.id)

    def trunk_deleted_precommit(self, resource, event, trunk_plugin, payload):
        trunk = payload.states[0]
        parent_port = payload.states[1]
        if self._is_port_bound(parent_port):
            raise trunk_exc.TrunkInUse(trunk_id=trunk.id)

    def subports_added(self, resource, event, trunk_plugin, payload):
        trunk = payload.states[0]
        subports = payload.metadata['subports']
        # Check if parent port is handled by OVN.
        if not self.plugin_driver.nb_ovn.lookup('Logical_Switch_Port',
                                                trunk.port_id, default=None):
            return
        if subports:
            self._set_sub_ports(trunk.port_id, subports)
        trunk.update(status=trunk_consts.TRUNK_ACTIVE_STATUS)

    def subports_deleted(self, resource, event, trunk_plugin, payload):
        trunk = payload.states[0]
        subports = payload.metadata['subports']
        # Check if parent port is handled by OVN.
        if not self.plugin_driver.nb_ovn.lookup('Logical_Switch_Port',
                                                trunk.port_id, default=None):
            return
        if subports:
            self._unset_sub_ports(subports)
        trunk.update(status=trunk_consts.TRUNK_ACTIVE_STATUS)


class OVNTrunkDriver(trunk_base.DriverBase):
    @property
    def is_loaded(self):
        try:
            return (ovn_const.OVN_ML2_MECH_DRIVER_NAME in
                    cfg.CONF.ml2.mechanism_drivers)
        except cfg.NoSuchOptError:
            return False

    @registry.receives(resources.TRUNK_PLUGIN, [events.AFTER_INIT])
    def register(self, resource, event, trigger, payload=None):
        super(OVNTrunkDriver, self).register(
            resource, event, trigger, payload=payload)
        self._handler = OVNTrunkHandler(self.plugin_driver)

        registry.subscribe(
            self._handler.trunk_created_precommit,
            resources.TRUNK,
            events.PRECOMMIT_CREATE)
        registry.subscribe(
            self._handler.trunk_deleted_precommit,
            resources.TRUNK,
            events.PRECOMMIT_DELETE)
        registry.subscribe(
            self._handler.trunk_created, resources.TRUNK, events.AFTER_CREATE)
        registry.subscribe(
            self._handler.trunk_deleted, resources.TRUNK, events.AFTER_DELETE)

        registry.subscribe(
            self._handler.subports_added,
            resources.SUBPORTS,
            events.AFTER_CREATE)
        registry.subscribe(
            self._handler.subports_deleted,
            resources.SUBPORTS,
            events.AFTER_DELETE)

    @classmethod
    def create(cls, plugin_driver):
        cls.plugin_driver = plugin_driver
        return cls(ovn_const.OVN_ML2_MECH_DRIVER_NAME,
                   SUPPORTED_INTERFACES,
                   SUPPORTED_SEGMENTATION_TYPES,
                   None,
                   can_trunk_bound_port=True)



# Copyright (c) 2014 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import eventlet

from neutron.common import utils
from neutron.notifiers import batch_notifier
from neutron.tests import base


class TestBatchNotifier(base.BaseTestCase):
    def setUp(self):
        super(TestBatchNotifier, self).setUp()
        self._received_events = eventlet.Queue()
        self.notifier = batch_notifier.BatchNotifier(2, self._queue_events)
        self.spawn_n_p = mock.patch.object(eventlet, 'spawn_n')

    def _queue_events(self, events):
        for event in events:
            self._received_events.put(event)

    def test_queue_event_no_event(self):
        spawn_n = self.spawn_n_p.start()
        self.notifier.queue_event(None)
        self.assertEqual(0, len(self.notifier._pending_events.queue))
        self.assertEqual(0, spawn_n.call_count)

    def test_queue_event_first_event(self):
        spawn_n = self.spawn_n_p.start()
        self.notifier.queue_event(mock.Mock())
        self.assertEqual(1, len(self.notifier._pending_events.queue))
        self.assertEqual(1, spawn_n.call_count)

    def test_queue_event_multiple_events_notify_method(self):
        def _batch_notifier_dequeue():
            while not self.notifier._pending_events.empty():
                self.notifier._pending_events.get()

        c_mock = mock.patch.object(self.notifier, '_notify',
                                   side_effect=_batch_notifier_dequeue).start()
        events = 20
        for i in range(events):
            self.notifier.queue_event('Event %s' % i)
            eventlet.sleep(0)  # yield to let coro execute

        utils.wait_until_true(self.notifier._pending_events.empty,
                              timeout=5)
        # Called twice: when the first thread calls "synced_send" and then,
        # in the same loop, when self._pending_events is not empty(). All
        # self.notifier.queue_event calls are done in just one
        # "batch_interval" (2 secs).
        self.assertEqual(2, c_mock.call_count)

    def test_queue_event_multiple_events_callback_method(self):
        events = 20
        for i in range(events):
            self.notifier.queue_event('Event %s' % i)
            eventlet.sleep(0)  # yield to let coro execute

        utils.wait_until_true(self.notifier._pending_events.empty,
                              timeout=5)
        expected = ['Event %s' % i for i in range(events)]
        # Check the events have been handled in the same input order.
        self.assertEqual(expected, list(self._received_events.queue))



# Copyright (c) 2019 Intel Corporation.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from neutron_lib.api.definitions import network_segment_range as apidef
from neutron_lib.api import extensions as api_extensions
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib.services import base as service_base
from oslo_log import log as logging

from neutron.api import extensions
from neutron.api.v2 import base

LOG = logging.getLogger(__name__)


class Network_segment_range(api_extensions.APIExtensionDescriptor):
    """Extension class supporting Network segment ranges.

    This class is used by neutron's extension framework to make metadata
    about the network segment range extension available to clients.

    With admin rights, one will be able to create, update, read and delete the
    values.
    """

    api_definition = apidef

    @classmethod
    def get_resources(cls):
        """Returns extension resources"""
        plugin = directory.get_plugin(plugin_constants.NETWORK_SEGMENT_RANGE)
        collection_name = apidef.COLLECTION_NAME.replace('_', '-')
        params = apidef.RESOURCE_ATTRIBUTE_MAP.get(apidef.COLLECTION_NAME,
                                                   dict())
        controller = base.create_resource(collection_name,
                                          apidef.RESOURCE_NAME,
                                          plugin, params, allow_bulk=True,
                                          allow_pagination=True,
                                          allow_sorting=True)

        ex = extensions.ResourceExtension(collection_name, controller,
                                          attr_map=params)

        return [ex]

    @classmethod
    def get_plugin_interface(cls):
        return NetworkSegmentRangePluginBase


class NetworkSegmentRangePluginBase(service_base.ServicePluginBase,
                                    metaclass=abc.ABCMeta):
    """REST API to manage network segment ranges.

    All methods must be in an admin context.
    """

    @classmethod
    def get_plugin_type(cls):
        return plugin_constants.NETWORK_SEGMENT_RANGE

    def get_plugin_description(self):
        return "Adds network segment ranges to Neutron resources"

    @abc.abstractmethod
    def create_network_segment_range(self, context, network_segment_range):
        """Create a network segment range.

        Create a network segment range, which represents the range of L2
        segments for tenant network allocation.

        :param context: neutron api request context
        :param network_segment_range: dictionary describing the network segment
            range, with keys as listed in the :obj:`RESOURCE_ATTRIBUTE_MAP`
            object in
            :file:`neutron_lib/api/definitions/network_segment_range.py`.
        """
        pass

    @abc.abstractmethod
    def delete_network_segment_range(self, context, id):
        """Delete a network segment range.

        :param context: neutron api request context
        :param id: UUID representing the network segment range to delete.
        """
        pass

    @abc.abstractmethod
    def update_network_segment_range(self, context, id, network_segment_range):
        """Update values of a network segment range.

        :param context: neutron api request context
        :param id: UUID representing the network segment range to update.
        :param network_segment_range: dictionary with keys indicating fields to
            update. valid keys are those that have a value of True for
            'allow_put' as listed in the :obj:`RESOURCE_ATTRIBUTE_MAP`
            object in
            :file:`neutron_lib/api/definitions/network_segment_range.py`.
        """
        pass

    @abc.abstractmethod
    def get_network_segment_ranges(self, context, filters=None, fields=None,
                                   sorts=None, limit=None, marker=None,
                                   page_reverse=False):
        """Retrieve a list of network segment ranges.

        The contents of the list depends on the filters.

        :param context: neutron api request context
        :param filters: a dictionary with keys that are valid keys for
                        a network segment range as listed in the
                        :obj:`RESOURCE_ATTRIBUTE_MAP` object in
                        :file:`neutron_lib/api/definitions/
                        network_segment_range.py`.
                        Values in this dictionary are an iterable containing
                        values that will be used for an exact match
                        comparison for that value. Each result returned by
                        this function will have matched one of the values
                        for each key in filters.
        :param fields: a list of strings that are valid keys in a
                       network segment range dictionary as listed in the
                       :obj:`RESOURCE_ATTRIBUTE_MAP` object in
                       :file:`neutron_lib/api/definitions/
                       network_segment_range.py`.
                       Only these fields will be returned.
        :param sorts: A list of (key, direction) tuples.
                      direction: True == ASC, False == DESC
        :param limit: maximum number of items to return
        :param marker: the last item of the previous page; when used, returns
                       next results after the marker resource.
        :param page_reverse: True if sort direction is reversed.
        """
        pass

    @abc.abstractmethod
    def get_network_segment_range(self, context, id, fields=None):
        """Retrieve a network segment range.

        :param context: neutron api request context
        :param id: UUID representing the network segment range to fetch.
        :param fields: a list of strings that are valid keys in a
                       network segment range dictionary as listed in the
                       :obj:`RESOURCE_ATTRIBUTE_MAP` object in
                       :file:`neutron_lib/api/definitions/
                       network_segment_range.py`.
                       Only these fields will be returned.
        """
        pass



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from os_ken.ofproto import ofproto_v1_3
from os_ken.ofproto import ofproto_v1_3_parser

from neutron.plugins.ml2.drivers.openvswitch.agent.openflow.native \
    import ofswitch
from neutron.tests import base


class FakeReply(object):
    def __init__(self, type):
        self.type = type


class TestBundledOpenFlowBridge(base.BaseTestCase):
    def setUp(self):
        super(TestBundledOpenFlowBridge, self).setUp()
        br = mock.Mock(spec=['install_instructions', 'foo'])
        br._get_dp = lambda: (mock.Mock(), ofproto_v1_3, ofproto_v1_3_parser)
        br.active_bundles = set()
        self.br = ofswitch.BundledOpenFlowBridge(br, False, False)

    def test_method_calls(self):
        self.br.install_instructions(dummy_arg=1)
        self.br.br.install_instructions.assert_called_once_with(dummy_arg=1)

    def test_illegal_method_calls(self):
        # With python3, this can be written as "with assertRaises..."
        try:
            self.br.uninstall_foo()
            self.fail("Expected an exception")
        except Exception as e:
            self.assertIsInstance(e, AttributeError)
        try:
            self.br.foo()
            self.fail("Expected an exception")
        except Exception as e:
            self.assertIsInstance(e, AttributeError)

    def test_normal_bundle_context(self):
        self.assertIsNone(self.br.active_bundle)
        self.br.br._send_msg = mock.Mock(side_effect=[
            FakeReply(ofproto_v1_3.ONF_BCT_OPEN_REPLY),
            FakeReply(ofproto_v1_3.ONF_BCT_COMMIT_REPLY)])
        with self.br:
            self.assertIsNotNone(self.br.active_bundle)
            # Do nothing
        # Assert that the active bundle is gone
        self.assertIsNone(self.br.active_bundle)

    def test_aborted_bundle_context(self):
        self.assertIsNone(self.br.active_bundle)
        self.br.br._send_msg = mock.Mock(side_effect=[
            FakeReply(ofproto_v1_3.ONF_BCT_OPEN_REPLY),
            FakeReply(ofproto_v1_3.ONF_BCT_DISCARD_REPLY)])
        try:
            with self.br:
                self.assertIsNotNone(self.br.active_bundle)
                raise Exception()
        except Exception:
            pass
        # Assert that the active bundle is gone
        self.assertIsNone(self.br.active_bundle)
        self.assertEqual(2, len(self.br.br._send_msg.mock_calls))
        args, kwargs = self.br.br._send_msg.call_args_list[0]
        self.assertEqual(ofproto_v1_3.ONF_BCT_OPEN_REQUEST,
                         args[0].type)
        args, kwargs = self.br.br._send_msg.call_args_list[1]
        self.assertEqual(ofproto_v1_3.ONF_BCT_DISCARD_REQUEST,
                         args[0].type)

    def test_bundle_context_with_error(self):
        self.assertIsNone(self.br.active_bundle)
        self.br.br._send_msg = mock.Mock(side_effect=[
            FakeReply(ofproto_v1_3.ONF_BCT_OPEN_REPLY),
            RuntimeError])
        try:
            with self.br:
                saved_bundle_id = self.br.active_bundle
                self.assertIsNotNone(self.br.active_bundle)
            self.fail("Expected an exception")
        except RuntimeError:
            pass
        # Assert that the active bundle is gone
        self.assertIsNone(self.br.active_bundle)
        self.assertIn(saved_bundle_id, self.br.br.active_bundles)

        self.assertEqual(2, len(self.br.br._send_msg.mock_calls))
        args, kwargs = self.br.br._send_msg.call_args_list[0]
        self.assertEqual(ofproto_v1_3.ONF_BCT_OPEN_REQUEST,
                         args[0].type)
        args, kwargs = self.br.br._send_msg.call_args_list[1]
        self.assertEqual(ofproto_v1_3.ONF_BCT_COMMIT_REQUEST,
                         args[0].type)



# Copyright (c) 2013 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.agent import topics
from neutron_lib.api.definitions import dvr
from neutron_lib.api.definitions import extraroute
from neutron_lib.api.definitions import extraroute_atomic
from neutron_lib.api.definitions import fip_pf_detail
from neutron_lib.api.definitions import fip_port_details
from neutron_lib.api.definitions import floatingip_pools
from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.api.definitions import l3_ext_gw_mode
from neutron_lib.api.definitions import l3_ext_ha_mode
from neutron_lib.api.definitions import l3_flavors
from neutron_lib.api.definitions import l3_port_ip_change_not_allowed
from neutron_lib.api.definitions import network_ha
from neutron_lib.api.definitions import qos_gateway_ip
from neutron_lib.api.definitions import \
    router_admin_state_down_before_update as r_admin_state_down_before_update
from neutron_lib.api.definitions import router_availability_zone
from neutron_lib import constants as n_const
from neutron_lib.db import resource_extend
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib import rpc as n_rpc
from neutron_lib.services import base as service_base
from oslo_config import cfg
from oslo_log import helpers as log_helpers
from oslo_log import log as logging
from oslo_utils import importutils

from neutron.api.rpc.agentnotifiers import l3_rpc_agent_api
from neutron.api.rpc.handlers import l3_rpc
from neutron.db import dns_db
from neutron.db import extraroute_db
from neutron.db import l3_dvr_ha_scheduler_db
from neutron.db import l3_dvrscheduler_db
from neutron.db import l3_fip_pools_db
from neutron.db import l3_fip_port_details
from neutron.db import l3_fip_qos
from neutron.db import l3_gateway_ip_qos
from neutron.db import l3_hamode_db
from neutron.db import l3_hascheduler_db
from neutron.db.models import l3 as l3_models
from neutron.quota import resource_registry
from neutron import service
from neutron.services.l3_router.service_providers import driver_controller


LOG = logging.getLogger(__name__)


def disable_dvr_extension_by_config(aliases):
    if not cfg.CONF.enable_dvr:
        LOG.info('Disabled DVR extension.')
        if 'dvr' in aliases:
            aliases.remove('dvr')
        if r_admin_state_down_before_update.ALIAS in aliases:
            aliases.remove(r_admin_state_down_before_update.ALIAS)


def disable_l3_qos_extension_by_plugins(ext, aliases):
    qos_class = 'neutron.services.qos.qos_plugin.QoSPlugin'
    if all(p not in cfg.CONF.service_plugins for p in ['qos', qos_class]):
        if ext in aliases:
            aliases.remove(ext)


@resource_extend.has_resource_extenders
class L3RouterPlugin(service_base.ServicePluginBase,
                     extraroute_db.ExtraRoute_db_mixin,
                     l3_hamode_db.L3_HA_NAT_db_mixin,
                     l3_gateway_ip_qos.L3_gw_ip_qos_db_mixin,
                     l3_dvr_ha_scheduler_db.L3_DVR_HA_scheduler_db_mixin,
                     dns_db.DNSDbMixin,
                     l3_fip_qos.FloatingQoSDbMixin,
                     l3_fip_port_details.Fip_port_details_db_mixin,
                     l3_fip_pools_db.FloatingIPPoolsMixin):

    """Implementation of the Neutron L3 Router Service Plugin.

    This class implements a L3 service plugin that provides
    router and floatingip resources and manages associated
    request/response.
    All DB related work is implemented in classes
    l3_db.L3_NAT_db_mixin, l3_hamode_db.L3_HA_NAT_db_mixin,
    l3_dvr_db.L3_NAT_with_dvr_db_mixin, and extraroute_db.ExtraRoute_db_mixin.
    """
    _supported_extension_aliases = [dvr.ALIAS, l3_apidef.ALIAS,
                                    l3_ext_gw_mode.ALIAS,
                                    extraroute.ALIAS,
                                    extraroute_atomic.ALIAS,
                                    n_const.L3_AGENT_SCHEDULER_EXT_ALIAS,
                                    l3_ext_ha_mode.ALIAS,
                                    router_availability_zone.ALIAS,
                                    l3_flavors.ALIAS, "qos-fip",
                                    fip_port_details.ALIAS,
                                    fip_pf_detail.ALIAS,
                                    floatingip_pools.ALIAS,
                                    qos_gateway_ip.ALIAS,
                                    l3_port_ip_change_not_allowed.ALIAS,
                                    r_admin_state_down_before_update.ALIAS,
                                    network_ha.ALIAS,
                                    ]

    __native_pagination_support = True
    __native_sorting_support = True
    __filter_validation_support = True

    IP_UPDATE_NOT_ALLOWED_LIST = [
        n_const.DEVICE_OWNER_ROUTER_INTF,
        n_const.DEVICE_OWNER_ROUTER_HA_INTF,
        n_const.DEVICE_OWNER_HA_REPLICATED_INT,
        n_const.DEVICE_OWNER_ROUTER_SNAT,
        n_const.DEVICE_OWNER_DVR_INTERFACE]

    @resource_registry.tracked_resources(router=l3_models.Router,
                                         floatingip=l3_models.FloatingIP)
    def __init__(self):
        self.router_scheduler = importutils.import_object(
            cfg.CONF.router_scheduler_driver)
        self.add_periodic_l3_agent_status_check()
        super(L3RouterPlugin, self).__init__()
        if 'dvr' in self.supported_extension_aliases:
            l3_dvrscheduler_db.subscribe()
        if 'l3-ha' in self.supported_extension_aliases:
            l3_hascheduler_db.subscribe()
        self.agent_notifiers.update(
            {n_const.AGENT_TYPE_L3: l3_rpc_agent_api.L3AgentNotifyAPI()})

        rpc_worker = service.RpcWorker([self], worker_process_count=0)

        self.add_worker(rpc_worker)
        self.l3_driver_controller = driver_controller.DriverController(self)

    @property
    def supported_extension_aliases(self):
        if not hasattr(self, '_aliases'):
            aliases = self._supported_extension_aliases[:]
            disable_dvr_extension_by_config(aliases)
            disable_l3_qos_extension_by_plugins('qos-fip', aliases)
            disable_l3_qos_extension_by_plugins('qos-gateway-ip', aliases)
            self._aliases = aliases
        return self._aliases

    @log_helpers.log_method_call
    def start_rpc_listeners(self):
        # RPC support
        self.topic = topics.L3PLUGIN
        self.conn = n_rpc.Connection()
        self.endpoints = [l3_rpc.L3RpcCallback()]
        self.conn.create_consumer(self.topic, self.endpoints,
                                  fanout=False)
        return self.conn.consume_in_threads()

    @classmethod
    def get_plugin_type(cls):
        return plugin_constants.L3

    def get_plugin_description(self):
        """returns string description of the plugin."""
        return ("L3 Router Service Plugin for basic L3 forwarding"
                " between (L2) Neutron networks and access to external"
                " networks via a NAT gateway.")

    def router_supports_scheduling(self, context, router_id):
        return self.l3_driver_controller.uses_scheduler(context, router_id)

    def create_floatingip(self, context, floatingip):
        """Create floating IP.

        :param context: Neutron request context
        :param floatingip: data for the floating IP being created
        :returns: A floating IP object on success

        As the l3 router plugin asynchronously creates floating IPs
        leveraging the l3 agent, the initial status for the floating
        IP object will be DOWN.
        """
        return super(L3RouterPlugin, self).create_floatingip(
            context, floatingip,
            initial_status=n_const.FLOATINGIP_STATUS_DOWN)

    @staticmethod
    @resource_extend.extends([l3_apidef.ROUTERS])
    def add_flavor_id(router_res, router_db):
        router_res['flavor_id'] = router_db['flavor_id']



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api.definitions import port_security
from neutron_lib.objects import common_types
from oslo_versionedobjects import fields as obj_fields

from neutron.objects import base


class _PortSecurity(base.NeutronDbObject):
    fields = {
        'id': common_types.UUIDField(),
        'port_security_enabled': obj_fields.BooleanField(
            default=port_security.DEFAULT_PORT_SECURITY),
    }

    foreign_keys = {
        'Port': {'id': 'id'},
        'Network': {'id': 'id'},
    }



# Copyright 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import datetime

from oslo_utils import uuidutils

from neutron.agent.common import resource_processing_queue as queue
from neutron.tests import base

_uuid = uuidutils.generate_uuid
FAKE_ID = _uuid()
FAKE_ID_2 = _uuid()

PRIORITY_RPC = 0


class TestExclusiveResourceProcessor(base.BaseTestCase):

    def test_i_am_primary(self):
        primary = queue.ExclusiveResourceProcessor(FAKE_ID)
        not_primary = queue.ExclusiveResourceProcessor(FAKE_ID)
        primary_2 = queue.ExclusiveResourceProcessor(FAKE_ID_2)
        not_primary_2 = queue.ExclusiveResourceProcessor(FAKE_ID_2)

        self.assertTrue(primary._i_am_primary())
        self.assertFalse(not_primary._i_am_primary())
        self.assertTrue(primary_2._i_am_primary())
        self.assertFalse(not_primary_2._i_am_primary())

        primary.__exit__(None, None, None)
        primary_2.__exit__(None, None, None)

    def test_primary(self):
        primary = queue.ExclusiveResourceProcessor(FAKE_ID)
        not_primary = queue.ExclusiveResourceProcessor(FAKE_ID)
        primary_2 = queue.ExclusiveResourceProcessor(FAKE_ID_2)
        not_primary_2 = queue.ExclusiveResourceProcessor(FAKE_ID_2)

        self.assertEqual(primary, primary._primary)
        self.assertEqual(primary, not_primary._primary)
        self.assertEqual(primary_2, primary_2._primary)
        self.assertEqual(primary_2, not_primary_2._primary)

        primary.__exit__(None, None, None)
        primary_2.__exit__(None, None, None)

    def test__enter__(self):
        self.assertNotIn(FAKE_ID, queue.ExclusiveResourceProcessor._primaries)
        primary = queue.ExclusiveResourceProcessor(FAKE_ID)
        primary.__enter__()
        self.assertIn(FAKE_ID, queue.ExclusiveResourceProcessor._primaries)
        primary.__exit__(None, None, None)

    def test__exit__(self):
        primary = queue.ExclusiveResourceProcessor(FAKE_ID)
        not_primary = queue.ExclusiveResourceProcessor(FAKE_ID)
        primary.__enter__()
        self.assertIn(FAKE_ID, queue.ExclusiveResourceProcessor._primaries)
        not_primary.__enter__()
        not_primary.__exit__(None, None, None)
        self.assertIn(FAKE_ID, queue.ExclusiveResourceProcessor._primaries)
        primary.__exit__(None, None, None)
        self.assertNotIn(FAKE_ID, queue.ExclusiveResourceProcessor._primaries)

    def test_data_fetched_since(self):
        primary = queue.ExclusiveResourceProcessor(FAKE_ID)
        self.assertEqual(datetime.datetime.min,
                         primary._get_resource_data_timestamp())

        ts1 = datetime.datetime.utcnow() - datetime.timedelta(seconds=10)
        ts2 = datetime.datetime.utcnow()

        primary.fetched_and_processed(ts2)
        self.assertEqual(ts2, primary._get_resource_data_timestamp())
        primary.fetched_and_processed(ts1)
        self.assertEqual(ts2, primary._get_resource_data_timestamp())

        primary.__exit__(None, None, None)

    def test_updates(self):
        primary = queue.ExclusiveResourceProcessor(FAKE_ID)
        not_primary = queue.ExclusiveResourceProcessor(FAKE_ID)

        primary.queue_update(queue.ResourceUpdate(FAKE_ID, 0))
        not_primary.queue_update(queue.ResourceUpdate(FAKE_ID, 0))

        for update in not_primary.updates():
            raise Exception("Only the primary should process a resource")

        self.assertEqual(2, len(list(primary.updates())))

    def test_hit_retry_limit(self):
        tries = 1
        rpqueue = queue.ResourceProcessingQueue()
        update = queue.ResourceUpdate(FAKE_ID, PRIORITY_RPC, tries=tries)
        rpqueue.add(update)
        self.assertFalse(update.hit_retry_limit())
        rpqueue.add(update)
        self.assertTrue(update.hit_retry_limit())



# Copyright 2015 Huawei Technologies India Pvt Ltd, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import sys

from neutron_lib import constants
from neutron_lib.objects import common_types
from neutron_lib.services.qos import constants as qos_consts
from neutron_lib.utils import helpers
from oslo_utils import versionutils
from oslo_versionedobjects import exception
from oslo_versionedobjects import fields as obj_fields

from neutron.db.qos import models as qos_db_model
from neutron.objects import base

DSCP_MARK = 'dscp_mark'


def get_rules(obj_cls, context, qos_policy_id):
    all_rules = []
    if not qos_policy_id:
        return all_rules

    with obj_cls.db_context_reader(context):
        for rule_type in qos_consts.VALID_RULE_TYPES:
            rule_cls_name = 'Qos%sRule' % helpers.camelize(rule_type)
            rule_cls = getattr(sys.modules[__name__], rule_cls_name)

            rules = rule_cls.get_objects(context, qos_policy_id=qos_policy_id)
            all_rules.extend(rules)
    return all_rules


class QosRule(base.NeutronDbObject, metaclass=abc.ABCMeta):
    # Version 1.0: Initial version, only BandwidthLimitRule
    #         1.1: Added DscpMarkingRule
    #         1.2: Added QosMinimumBandwidthRule
    #         1.3: Added direction for BandwidthLimitRule
    #         1.4: Added PacketRateLimitRule
    #         1.5: Added QosMinimumPacketRateRule
    #
    # NOTE(mangelajo): versions need to be handled from the top QosRule object
    #                  because it's the only reference QosPolicy can make
    #                  to them via obj_relationships version map
    VERSION = '1.5'

    fields = {
        'id': common_types.UUIDField(),
        'qos_policy_id': common_types.UUIDField()
    }

    fields_no_update = ['id', 'qos_policy_id']

    # should be redefined in subclasses
    rule_type = None

    duplicates_compare_fields = ()

    def duplicates(self, other_rule):
        """Returns True if rules have got same values in fields defined in
        'duplicates_compare_fields' list.

        In case when subclass don't have defined any field in
        duplicates_compare_fields, only rule types are compared.
        """

        if self.rule_type != other_rule.rule_type:
            return False

        if self.duplicates_compare_fields:
            for field in self.duplicates_compare_fields:
                if getattr(self, field) != getattr(other_rule, field):
                    return False
        return True

    def to_dict(self):
        dict_ = super(QosRule, self).to_dict()
        dict_['type'] = self.rule_type
        return dict_

    def should_apply_to_port(self, port):
        """Check whether a rule can be applied to a specific port.

        This function has the logic to decide whether a rule should
        be applied to a port or not, depending on the source of the
        policy (network, or port). Eventually rules could override
        this method, or we could make it abstract to allow different
        rule behaviour.
        """
        is_port_policy = self.qos_policy_id == port[qos_consts.QOS_POLICY_ID]
        is_network_policy_only = port[qos_consts.QOS_POLICY_ID] is None
        is_network_device_port = any(port['device_owner'].startswith(prefix)
                                     for prefix
                                     in constants.DEVICE_OWNER_PREFIXES)
        # NOTE(miouge): Network QoS policies should apply to ext routers ports:
        # - DEVICE_OWNER_AGENT_GW for DVR routers
        # - DEVICE_OWNER_ROUTER_GW for normal neutron routers
        is_router_gw = any(port['device_owner'].startswith(prefix)
                           for prefix in [constants.DEVICE_OWNER_AGENT_GW,
                                          constants.DEVICE_OWNER_ROUTER_GW])
        # NOTE(ralonsoh): return True if:
        #    - Is a port QoS policy (not a network QoS policy)
        #    - Is not an internal network device (e.g. router) and is a network
        #      QoS policy and there is no port QoS policy
        return (is_port_policy or
                ((is_router_gw or not is_network_device_port) and
                 is_network_policy_only))

    def obj_make_compatible(self, primitive, target_version):
        _target_version = versionutils.convert_version_to_tuple(target_version)
        if _target_version < (1, 3):
            raise exception.IncompatibleObjectVersion(
                objver=target_version, objtype=self.__class__.__name__)


@base.NeutronObjectRegistry.register
class QosBandwidthLimitRule(QosRule):

    db_model = qos_db_model.QosBandwidthLimitRule

    fields = {
        'max_kbps': obj_fields.IntegerField(nullable=True),
        'max_burst_kbps': obj_fields.IntegerField(nullable=True),
        'direction': common_types.FlowDirectionEnumField(
            default=constants.EGRESS_DIRECTION)
    }

    duplicates_compare_fields = ['direction']

    rule_type = qos_consts.RULE_TYPE_BANDWIDTH_LIMIT


@base.NeutronObjectRegistry.register
class QosDscpMarkingRule(QosRule):

    db_model = qos_db_model.QosDscpMarkingRule

    fields = {
        DSCP_MARK: common_types.DscpMarkField(),
    }

    rule_type = qos_consts.RULE_TYPE_DSCP_MARKING


@base.NeutronObjectRegistry.register
class QosMinimumBandwidthRule(QosRule):

    db_model = qos_db_model.QosMinimumBandwidthRule

    fields = {
        'min_kbps': obj_fields.IntegerField(nullable=True),
        'direction': common_types.FlowDirectionEnumField(),
    }

    duplicates_compare_fields = ['direction']

    rule_type = qos_consts.RULE_TYPE_MINIMUM_BANDWIDTH


@base.NeutronObjectRegistry.register
class QosPacketRateLimitRule(QosRule):

    db_model = qos_db_model.QosPacketRateLimitRule

    fields = {
        'max_kpps': obj_fields.IntegerField(nullable=True),
        'max_burst_kpps': obj_fields.IntegerField(nullable=True),
        'direction': common_types.FlowDirectionEnumField(
            default=constants.EGRESS_DIRECTION)
    }

    duplicates_compare_fields = ['direction']

    rule_type = qos_consts.RULE_TYPE_PACKET_RATE_LIMIT


@base.NeutronObjectRegistry.register
class QosMinimumPacketRateRule(QosRule):

    db_model = qos_db_model.QosMinimumPacketRateRule

    fields = {
        'min_kpps': obj_fields.IntegerField(nullable=False),
        'direction': common_types.FlowDirectionAndAnyEnumField(),
    }

    duplicates_compare_fields = ['direction']

    rule_type = qos_consts.RULE_TYPE_MINIMUM_PACKET_RATE
